"""
Tests for functionalities in orthoseg.train.
"""

import os
import shutil
import tempfile
from datetime import datetime
from pathlib import Path

import gdown
import geofileops as gfo
import pytest

import orthoseg
import orthoseg.model.model_helper as mh
from orthoseg.helpers import config_helper as conf
from tests import test_helper

testprojects_dir = Path(tempfile.gettempdir()) / "orthoseg_test_end2end/sample_projects"
footballfields_dir = testprojects_dir / "footballfields"
projecttemplate_dir = testprojects_dir / "project_template"


def get_testdata_dir() -> Path:
    return Path(__file__).resolve().parent / "data"


def test_1_init_testproject():
    # Use footballfields sample project for these end to end tests
    shutil.rmtree(testprojects_dir, ignore_errors=True)
    shutil.copytree(test_helper.sampleprojects_dir, testprojects_dir)


@pytest.mark.order(after="test_1_init_testproject")
def test_2_load_images():
    # Load project config to init some vars.
    config_path = footballfields_dir / "footballfields_BEFL-2019_test.ini"
    conf.read_orthoseg_config(config_path)
    image_cache_dir = conf.dirs.getpath("predict_image_input_dir")

    # Clean result if it isn't empty yet
    if image_cache_dir.exists():
        shutil.rmtree(image_cache_dir)
        # Make sure is is deleted now!
        assert not image_cache_dir.exists()

    # Run task to load images
    orthoseg.load_images(config_path)

    # Check if the right number of files was loaded
    assert image_cache_dir.exists()
    files = list(image_cache_dir.glob("**/*.jpg"))
    assert len(files) == 6


@pytest.mark.skipif(
    "GITHUB_ACTIONS" in os.environ and os.name == "nt",
    reason="crashes on github CI on windows",
)
@pytest.mark.order(after="test_1_init_testproject")
def test_3_train():
    # Load project config to init some vars.
    config_path = footballfields_dir / "footballfields_train_test.ini"
    conf.read_orthoseg_config(config_path)

    # Init + cleanup result dirs
    traindata_id_result = 2
    training_dir = conf.dirs.getpath("training_dir")
    training_id_dir = training_dir / f"{traindata_id_result:02d}"
    if training_id_dir.exists():
        shutil.rmtree(training_id_dir)
    model_dir = conf.dirs.getpath("model_dir")
    if model_dir.exists():
        modelfile_paths = model_dir.glob(f"footballfields_{traindata_id_result:02d}_*")
        for modelfile_path in modelfile_paths:
            modelfile_path.unlink()

    # Make sure the label files in version 01 are older than those in the label dir
    # so a new model will be trained
    label_01_path = training_dir / "01/footballfields_BEFL-2019_polygons.gpkg"
    timestamp_old = datetime(year=2020, month=1, day=1).timestamp()
    os.utime(label_01_path, (timestamp_old, timestamp_old))

    # Run train session
    orthoseg.train(config_path)

    # Check if the training (image) data was created
    assert training_id_dir.exists()

    # Check if the new model was created
    best_model = mh.get_best_model(
        model_dir=conf.dirs.getpath("model_dir"),
        segment_subject=conf.general["segment_subject"],
        traindata_id=2,
    )

    assert best_model is not None
    assert best_model["traindata_id"] == traindata_id_result
    assert best_model["epoch"] == 0


@pytest.mark.skipif(
    "GITHUB_ACTIONS" in os.environ and os.name == "nt",
    reason="crashes on github CI on windows",
)
@pytest.mark.order(after="test_2_load_images")
def test_4_predict():
    # Load project config to init some vars.
    config_path = footballfields_dir / "footballfields_BEFL-2019_test.ini"
    conf.read_orthoseg_config(config_path)

    # Cleanup result if it isn't empty yet
    predict_image_output_basedir = conf.dirs.getpath("predict_image_output_basedir")
    predict_image_output_dir = (
        predict_image_output_basedir.parent
        / f"{predict_image_output_basedir.name}_footballfields_02_0"
    )
    if predict_image_output_dir.exists():
        shutil.rmtree(predict_image_output_dir)
        # Make sure it is deleted now!
        assert not predict_image_output_dir.exists()
    result_vector_dir = conf.dirs.getpath("output_vector_dir")
    if result_vector_dir.exists():
        shutil.rmtree(result_vector_dir)
        # Make sure is is deleted now!
        assert not result_vector_dir.exists()

    # Download the version 01 model
    model_dir = conf.dirs.getpath("model_dir")
    model_dir.mkdir(parents=True, exist_ok=True)
    model_hdf5_path = model_dir / "footballfields_01_0.97392_201.hdf5"
    if not model_hdf5_path.exists():
        gdown.download(
            id="1UlNorZ74ADCr3pL4MCJ_tnKRNoeZX79g", output=str(model_hdf5_path)
        )
    model_hyperparams_path = model_dir / "footballfields_01_hyperparams.json"
    if not model_hyperparams_path.exists():
        gdown.download(
            id="1NwrVVjx9IsjvaioQ4-bkPMrq7S6HeWIo", output=str(model_hyperparams_path)
        )
    model_modeljson_path = model_dir / "footballfields_01_model.json"
    if not model_modeljson_path.exists():
        gdown.download(
            id="1LNPLypM5in3aZngBKK_U4Si47Oe97ZWN", output=str(model_modeljson_path)
        )

    # Run task to predict
    orthoseg.predict(config_path)

    # Check results
    result_vector_path = result_vector_dir / "footballfields_01_201_BEFL-2019.gpkg"
    assert result_vector_path.exists()
    result_gdf = gfo.read_file(result_vector_path)
    if os.name == "nt":
        assert len(result_gdf) == 211
    else:
        assert len(result_gdf) == 211


@pytest.mark.skipif(
    "GITHUB_ACTIONS" in os.environ and os.name == "nt",
    reason="crashes on github CI on windows",
)
@pytest.mark.order(after="test_4_predict")
def test_5_postprocess():
    # Load project config to init some vars.
    config_path = footballfields_dir / "footballfields_BEFL-2019_test.ini"
    conf.read_orthoseg_config(config_path)

    # Cleanup result if it isn't empty yet
    result_vector_dir = conf.dirs.getpath("output_vector_dir")
    result_diss_path = (
        result_vector_dir / "footballfields_01_201_BEFL-2019_dissolve.gpkg"
    )
    if result_diss_path.exists():
        gfo.remove(result_diss_path)

    # Run task to postprocess
    orthoseg.postprocess(config_path)

    # Check results
    assert result_diss_path.exists()
    result_gdf = gfo.read_file(result_diss_path)
    if os.name == "nt":
        assert len(result_gdf) == 207
    else:
        assert len(result_gdf) == 207



"""Module to prepare the training datasets."""

import logging
import math
import pprint
import shutil
import warnings
from pathlib import Path
from typing import Optional, Union

import geofileops as gfo
import geopandas as gpd
import numpy as np
import pandas as pd
import rasterio as rio
import rasterio.features as rio_features
import rasterio.profiles as rio_profiles
import shapely
import shapely.geometry as sh_geom
from PIL import Image

from orthoseg.util import ows_util, vector_util
from orthoseg.util.progress_util import ProgressLogger

# Get a logger...
logger = logging.getLogger(__name__)


class LabelInfo:
    """Information needed to find train labels."""

    def __init__(
        self,
        locations_path: Path,
        polygons_path: Path,
        image_layer: str,
        pixel_x_size: Optional[float] = None,
        pixel_y_size: Optional[float] = None,
        locations_gdf: Optional[gpd.GeoDataFrame] = None,
        polygons_gdf: Optional[gpd.GeoDataFrame] = None,
    ):
        """Conctructor of LabelInfo.

        Args:
            locations_path (Path): file path to the locations file.
            polygons_path (Path): file path to the polygons file.
            image_layer (str): image layer for the label.
            pixel_x_size (Optional[float], optional): _description_. Defaults to None.
            pixel_y_size (Optional[float], optional): _description_. Defaults to None.
            locations_gdf (Optional[gpd.GeoDataFrame], optional): _description_.
                Defaults to None.
            polygons_gdf (Optional[gpd.GeoDataFrame], optional): _description_.
                Defaults to None.
        """
        self.locations_path = locations_path
        self.polygons_path = polygons_path
        self.image_layer = image_layer
        self.pixel_x_size = pixel_x_size
        self.pixel_y_size = pixel_y_size
        self.locations_gdf = locations_gdf
        self.polygons_gdf = polygons_gdf

    def __repr__(self) -> str:
        """Return the object information as readable string.

        Returns:
            str: readable string representation of object.
        """
        repr = (
            f"LabelInfo(image_layer={self.image_layer}, locations_path="
            f"{self.locations_path}, polygons_path={self.polygons_path},...)"
        )
        return repr


class ValidationError(ValueError):
    """A validation exception.

    Args:
        ValueError (_type_): _description_
    """

    def __init__(self, message, errors):
        """Conctructor of ValidationError.

        Args:
            message (str): error message.
            errors (list[str]): List of validation errors.
        """
        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.errors = errors

    def __repr__(self):
        """Formats validation errors to string.

        Returns:
            str: returns the validation errors as string.
        """
        repr = super().__repr__()
        if self.errors is not None and len(self.errors) > 0:
            repr += f"\n  -> Errors: {pprint.pformat(self.errors, indent=4, width=100)}"
        return repr

    def __str__(self) -> str:
        """Formats validation errors to string.

        Returns:
            str: returns the validation errors as string.
        """
        repr = super().__str__()
        if self.errors is not None and len(self.errors) > 0:
            repr += f"\n  -> Errors: {pprint.pformat(self.errors, indent=4, width=100)}"
        return repr

    def to_html(self) -> str:
        """Formats validation errors to html.

        Returns:
            str: the validation errors as html.
        """
        repr = super().__str__()
        if self.errors is not None and len(self.errors) > 0:
            errors_df = pd.DataFrame(self.errors, columns=["error"])
            repr += f"{errors_df.to_html()}"
        return repr


def prepare_traindatasets(
    label_infos: list[LabelInfo],
    classes: dict,
    image_layers: dict,
    training_dir: Path,
    labelname_column: str = "classname",
    image_pixel_x_size: float = 0.25,
    image_pixel_y_size: float = 0.25,
    image_pixel_width: int = 512,
    image_pixel_height: int = 512,
    ssl_verify: Union[bool, str] = True,
    only_validate: bool = False,
    force: bool = False,
) -> tuple[Path, int]:
    """This function prepares training data for the vector labels provided.

    It will:
        * get orthophoto images from the correct image_layer
        * create the corresponding label mask for each image

    Returns a tuple with (output_dir, dataversion):
        output_dir: the dir where the traindataset was created/found
        dataversion: a version number for the dataset created/found

    Args:
        label_infos (list[LabelInfo]): paths to the files with label polygons
            and locations to generate images for.
        classes (dict): dict with the classes to detect as keys. The values
            are the following:
                - labelnames: list of labels to use for this class
                - weight:
                - burn_value:
        image_layers (dict): the image layers available with their properties.
        training_dir (Path): the directory to save the training data to.
        labelname_column (str): the column where the label names are stored in
            the polygon files. If the column name specified is not found, column
            "label_name" is used if it exists for backwards compatibility.
        image_pixel_x_size (float): pixel size in x direction. Defaults to 0.25.
        image_pixel_y_size (float): pixel size in y direction. Defaults to 0.25.
        image_pixel_width (int): width of the image in pixels. Defaults to 512.
        image_pixel_height (int): height of the image in pixels. Defaults to 512.
        ssl_verify (bool or str, optional): True to use the default
            certificate bundle as installed on your system. False disables
            certificate validation (NOT recommended!). If a path to a
            certificate bundle file (.pem) is passed, this will be used.
            In corporate networks using a proxy server this is often needed
            to evade CERTIFICATE_VERIFY_FAILED errors. Defaults to True.
        only_validate (bool, optional): True to only validate the input label data.
        force (bool, opitional): True to force recreation of output files.
            Defaults to False.
    """
    # Check if the first class is named "background"
    if len(classes) == 0:
        raise Exception("No classes specified")
    elif next(iter(classes)).lower() != "background":
        classes_str = pprint.pformat(classes, sort_dicts=False, width=50)
        raise Exception(
            f"By convention, the first class must be called background!\n{classes_str}"
        )

    # Check if the latest version of training data is already ok
    # Determine the current data version based on existing output data dir(s),
    # If dir ends on _TMP_* ignore it, as it (probably) ended with an error.
    output_dirs = [
        output_dir
        for output_dir in training_dir.glob("[0-9]*/")
        if "_TMP_" not in output_dir.name
    ]

    reuse_traindata = False
    dataversion_mostrecent = None
    if len(output_dirs) == 0:
        dataversion_new = 1
    else:
        # Get the output dir with the highest version (=first if sorted desc)
        output_dir_mostrecent = sorted(output_dirs, reverse=True)[0]
        dataversion_mostrecent = int(output_dir_mostrecent.name)

        # If none of the input files changed since previous run, reuse dataset
        for label_file in label_infos:
            reuse_traindata = True
            labellocations_output_mostrecent_path = (
                output_dir_mostrecent / label_file.locations_path.name
            )
            labeldata_output_mostrecent_path = (
                output_dir_mostrecent / label_file.polygons_path.name
            )
            if not (
                labellocations_output_mostrecent_path.exists()
                and labeldata_output_mostrecent_path.exists()
                and gfo.cmp(
                    label_file.locations_path, labellocations_output_mostrecent_path
                )
                and gfo.cmp(label_file.polygons_path, labeldata_output_mostrecent_path)
            ):
                reuse_traindata = False
                break
        if reuse_traindata:
            dataversion_new = dataversion_mostrecent
            logger.info(
                "Input label file(s) didn't change since last prepare_traindatasets, "
                f"so reuse version {dataversion_new}"
            )
        else:
            dataversion_new = dataversion_mostrecent + 1
            logger.info(
                "Input label file(s) changed since last prepare_traindatasets, "
                f"so create new training data version {dataversion_new}"
            )

    # Determine the output dir
    training_dataversion_dir = training_dir / f"{dataversion_new:02d}"

    # If the train data is already ok, just return
    if reuse_traindata is True:
        return (training_dataversion_dir, dataversion_new)

    # The input labels have changed: create new train version
    # -------------------------------------------------------
    # read, validate and prepare data
    labeldata = prepare_labeldata(
        label_infos=label_infos,
        classes=classes,
        labelname_column=labelname_column,
        image_pixel_x_size=image_pixel_x_size,
        image_pixel_y_size=image_pixel_y_size,
        image_pixel_width=image_pixel_width,
        image_pixel_height=image_pixel_height,
    )

    # Reading label data was succesfull.
    if only_validate:
        return (training_dataversion_dir, dataversion_new)

    # Prepare temp dir to put training dataset in.
    # Use a temp dir, so it can be removed/ignored if an error occurs later on.
    output_tmp_dir = create_tmp_dir(
        training_dir, f"{dataversion_new:02d}", remove_existing=True
    )

    # Copy the label input files to dest dir + to a backup dir.
    for label_file in label_infos:
        gfo.copy(label_file.locations_path, output_tmp_dir)
        gfo.copy(label_file.polygons_path, output_tmp_dir)
        backup_dir = label_file.locations_path.parent / f"backup_v{dataversion_new:02d}"
        shutil.rmtree(backup_dir, ignore_errors=True)
        backup_dir.mkdir(parents=True, exist_ok=True)
        gfo.copy(label_file.locations_path, backup_dir)
        gfo.copy(label_file.polygons_path, backup_dir)

    # Now create the images/masks for the new train version for the different traindata
    # types.
    traindata_types = ["train", "validation", "test"]
    nb_todo = 0
    for labellocations_gdf, _ in labeldata:
        labellocations_curr_gdf = labellocations_gdf[
            labellocations_gdf["traindata_type"].isin(traindata_types)
        ]
        nb_todo += len(labellocations_curr_gdf)
    progress = ProgressLogger(message="prepare training images", nb_steps_total=nb_todo)
    logger.info(f"Get images for {nb_todo} labels")

    for traindata_type in traindata_types:
        if dataversion_mostrecent is not None:
            previous_dataversion_dir = training_dir / f"{dataversion_mostrecent:02d}"
            previous_imagedata_image_dir = (
                previous_dataversion_dir / traindata_type / "image"
            )
        # Create output dirs...
        output_imagedatatype_dir = output_tmp_dir / traindata_type
        output_imagedata_image_dir = output_imagedatatype_dir / "image"
        output_imagedata_mask_dir = output_imagedatatype_dir / "mask"
        for dir in [
            output_imagedatatype_dir,
            output_imagedata_mask_dir,
            output_imagedata_image_dir,
        ]:
            if dir and not dir.exists():
                dir.mkdir(parents=True, exist_ok=True)

        for labellocations_gdf, labels_to_burn_gdf in labeldata:
            try:
                # Get the label locations for this traindata type
                labellocations_curr_gdf = labellocations_gdf[
                    labellocations_gdf["traindata_type"] == traindata_type
                ]

                # Loop trough all locations labels to get an image for each of them
                for i, label_tuple in enumerate(labellocations_curr_gdf.itertuples()):
                    img_bbox = label_tuple.geometry
                    image_layer = getattr(label_tuple, "image_layer")

                    # Prepare file name for the image
                    assert labellocations_gdf.crs is not None
                    output_filename = ows_util.create_filename(
                        crs=labellocations_gdf.crs,
                        bbox=img_bbox.bounds,
                        size=(image_pixel_width, image_pixel_height),
                        image_format=ows_util.FORMAT_PNG,
                        layername="_".join(
                            image_layers[image_layer]["layersources"][0].layernames
                        ),
                    )

                    # If the image exists already in the previous version, reuse it.
                    if (
                        dataversion_mostrecent is not None
                        and (previous_imagedata_image_dir / output_filename).exists()
                    ):
                        image_filepath = shutil.copy(
                            src=previous_imagedata_image_dir / output_filename,
                            dst=output_imagedata_image_dir / output_filename,
                        )
                        pgw_filename = output_filename.replace(".png", ".pgw")
                        if (previous_imagedata_image_dir / pgw_filename).exists():
                            shutil.copy(
                                src=previous_imagedata_image_dir / pgw_filename,
                                dst=output_imagedata_image_dir / pgw_filename,
                            )
                    else:
                        # Get the image from the WMS service.
                        image_filepath = ows_util.getmap_to_file(
                            layersources=image_layers[image_layer]["layersources"],
                            output_dir=output_imagedata_image_dir,
                            crs=labellocations_gdf.crs,
                            bbox=img_bbox.bounds,
                            size=(image_pixel_width, image_pixel_height),
                            ssl_verify=ssl_verify,
                            image_format=ows_util.FORMAT_PNG,
                            # image_format_save=ows_util.FORMAT_TIFF,
                            image_pixels_ignore_border=image_layers[image_layer][
                                "image_pixels_ignore_border"
                            ],
                            transparent=False,
                            layername_in_filename=True,
                            output_filename=output_filename,
                        )

                    # Create a mask corresponding with the image file
                    # Mask should not be in a lossy format -> png!
                    mask_filepath = Path(
                        str(image_filepath)
                        .replace(
                            str(output_imagedata_image_dir),
                            str(output_imagedata_mask_dir),
                        )
                        .replace(".jpg", ".png")
                    )
                    nb_classes = len(classes)

                    # Only keep the labels that are meant for this image layer
                    if "image_layer" not in labels_to_burn_gdf.columns:
                        print("odd")
                    labels_for_layer_gdf = (
                        labels_to_burn_gdf.loc[
                            labels_to_burn_gdf["image_layer"] == image_layer
                        ]
                    ).copy()
                    if len(labels_for_layer_gdf) == 0:
                        logger.info(f"No polygons to burn for {image_layer=}!")
                    _create_mask(
                        input_image_filepath=image_filepath,
                        output_mask_filepath=mask_filepath,
                        labels_to_burn_gdf=labels_for_layer_gdf,
                        nb_classes=nb_classes,
                        force=force,
                    )

                    # Log the progress and prediction speed
                    progress.step()

            except Exception as ex:
                raise ex

    # If everything went fine, rename output_tmp_dir to the final output_dir
    output_tmp_dir.rename(training_dataversion_dir)

    return (training_dataversion_dir, dataversion_new)


def prepare_labeldata(
    label_infos: list[LabelInfo],
    classes: dict,
    labelname_column: str,
    image_pixel_x_size: Optional[float],
    image_pixel_y_size: Optional[float],
    image_pixel_width: int,
    image_pixel_height: int,
) -> list[tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]]:
    """Prepare and validate the data in the labelinfos.

    The GeoDataFrames returned can be used to fetch train images and burn masks.

    Args:
        label_infos (list[LabelInfo]): the label files/data.
        classes (dict): dict with classes and their corresponding
            label class names + weights.
        labelname_column (str): the column name in the label polygon files where the
            label classname can be found. Defaults to "classname".
        image_pixel_x_size (float): pixel size in x direction.
        image_pixel_y_size (float): pixel size in y direction.
        image_pixel_width (int): width of the image in pixels.
        image_pixel_height (int): height of the image in pixels.

    Raises:
        ValidationError: if the data is somehow not valid. All issues are listed in the
            errors property.

    Returns:
        list[tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]]: returns a list of tuples with
            geodataframes with labellocations and labelpolygons to burn.
    """
    labeldata_result = []
    validation_errors = []
    validation_warnings = []
    train_locations_found = False
    validation_locations_found = False
    for label_info in label_infos:
        # Read label locations if needed
        if label_info.locations_gdf is not None:
            labellocations_gdf = label_info.locations_gdf
        else:
            logger.debug(f"Read label locations from {label_info.locations_path}")
            labellocations_gdf = gfo.read_file(label_info.locations_path)
            labellocations_gdf["path"] = str(label_info.locations_path)
            labellocations_gdf["image_layer"] = label_info.image_layer
            # Remark: geopandas 0.7.0 drops fid column internally!
            labellocations_gdf["row_nb_orig"] = labellocations_gdf.index
            if len(labellocations_gdf) == 0:
                logger.warning(
                    f"No label locations found in {label_info.locations_path}"
                )
                continue

        # Read label polygons if needed
        if label_info.polygons_gdf is not None:
            labelpolygons_gdf = label_info.polygons_gdf
        else:
            logger.debug(f"Read label data from {label_info.polygons_path}")
            labelpolygons_gdf = gfo.read_file(label_info.polygons_path)
            labelpolygons_gdf["path"] = str(label_info.polygons_path)
            labelpolygons_gdf["image_layer"] = label_info.image_layer
            if len(labelpolygons_gdf) == 0:
                logger.warning(f"No label polygons found in {label_info.polygons_path}")

        assert labellocations_gdf is not None
        assert labelpolygons_gdf is not None

        if labellocations_gdf is None or len(labellocations_gdf) == 0:
            validation_errors.append("No label locations found in labellocations_gdf")
            continue

        # Validate + process the location data
        # ------------------------------------
        if "traindata_type" not in labellocations_gdf.columns:
            validation_errors.append(
                "Mandatory column traindata_type not found in "
                f"{label_info.locations_path}"
            )

        # Check + correct the geom of the location to make sure it matches the size of
        # the training images/masks to be generated...
        # Prepare pixel size: if defined on label info level, use that.
        pixel_x_size = label_info.pixel_x_size
        pixel_x_size = pixel_x_size if pixel_x_size is not None else image_pixel_x_size
        pixel_y_size = label_info.pixel_y_size
        pixel_y_size = pixel_y_size if pixel_y_size is not None else image_pixel_y_size
        assert pixel_x_size is not None
        assert pixel_y_size is not None

        # Tile width/height in units of crs
        image_crs_width = math.fabs(image_pixel_width * pixel_x_size)
        image_crs_height = math.fabs(image_pixel_height * pixel_y_size)

        locations_none = []
        for location in labellocations_gdf.itertuples():
            if location.geometry is None or location.geometry.is_empty:
                logger.warning(
                    f"No or empty geometry found in file {Path(location.path).name} "
                    f"for index {location.Index}, it will be ignored"
                )
                locations_none.append(location.Index)
                continue

            # Check if the traindata_type is valid
            if location.traindata_type not in ["train", "validation", "test", "todo"]:
                validation_errors.append(
                    f"Invalid traindata_type in {Path(location.path).name}: "
                    f"{location.geometry.wkt}"
                )

            # Check if the geometry is valid
            is_valid_reason = shapely.is_valid_reason(
                labellocations_gdf.geometry[location.Index]
            )
            if is_valid_reason != "Valid Geometry":
                validation_errors.append(
                    f"Invalid geometry in {Path(location.path).name}: "
                    f"{is_valid_reason}"
                )
                continue

            geom_bounds = location.geometry.bounds
            xmin = geom_bounds[0] - (geom_bounds[0] % image_pixel_x_size)
            ymin = geom_bounds[1] - (geom_bounds[1] % image_pixel_y_size)
            xmax = xmin + image_crs_width
            ymax = ymin + image_crs_height
            location_geom_aligned = sh_geom.box(xmin, ymin, xmax, ymax)

            # Check if the realigned geom overlaps good enough with the original
            intersection = location_geom_aligned.intersection(
                labellocations_gdf.at[location.Index, "geometry"]
            )
            assert image_pixel_x_size is not None
            assert image_pixel_y_size is not None
            area_1row_1col = (
                image_pixel_x_size * image_crs_width
                + image_pixel_y_size * image_crs_height
            )
            if intersection.area < (location_geom_aligned.area - area_1row_1col):
                # Original geom was digitized too small
                validation_errors.append(
                    f"Location geometry skewed or too small ({intersection.area}, "
                    f"based on train config expected {location_geom_aligned.area}) in "
                    f"{Path(location.path).name}: {location.geometry.wkt}"
                )
            elif sh_geom.box(*geom_bounds).area > location_geom_aligned.area * 1.1:
                validation_warnings.append(
                    f"Location geometry too large ({sh_geom.box(*geom_bounds).area}, "
                    f"based on train config expected {location_geom_aligned.area}) "
                    f"in file {Path(location.path).name}: {location.geometry.wkt}"
                )
            labellocations_gdf.at[location.Index, "geometry"] = location_geom_aligned

        # Remove locations with None or point/line geoms
        labellocations_gdf = labellocations_gdf[
            ~labellocations_gdf.index.isin(locations_none)
        ]

        # Check if labellocations has a proper crs
        if labellocations_gdf.crs is None:
            validation_errors.append(
                "No crs in labellocations, labellocation_gdf.crs: "
                f"{labellocations_gdf.crs}"
            )

        # Check that there is at least one train location and one validation location.
        if len(labellocations_gdf.query("traindata_type == 'train'")) > 0:
            train_locations_found = True
        if len(labellocations_gdf.query("traindata_type == 'validation'")) > 0:
            validation_locations_found = True

        # Validate + process the polygons data
        # ------------------------------------
        if labelpolygons_gdf is None:
            validation_errors.append("No labelpolygons in the training data!")
            continue

        # Create list with only the input polygons that need to be burned in the mask
        labels_to_burn_gdf = labelpolygons_gdf.copy()
        labels_to_burn_gdf["burn_value"] = None

        if labelname_column not in labelpolygons_gdf.columns:
            # For backwards compatibility, also support old default column name
            labelname_column = "label_name"

        if labelname_column in labelpolygons_gdf.columns:
            # If there is a column labelname_column, use the burn values specified in
            # the configuration
            for classname in classes:
                labels_to_burn_gdf.loc[
                    (
                        labels_to_burn_gdf[labelname_column].isin(
                            classes[classname]["labelnames"]
                        )
                    ),
                    "burn_value",
                ] = classes[classname]["burn_value"]

            # Check if there are invalid class names
            invalid_gdf = labels_to_burn_gdf.loc[
                labels_to_burn_gdf["burn_value"].isnull()
            ]
            for _, invalid_row in invalid_gdf.iterrows():
                validation_errors.append(
                    f"Invalid classname in {Path(invalid_row['path']).name}: "
                    f"{invalid_row[labelname_column]}"
                )

            # Filter away rows that are going to burn 0, as this is useless...
            labels_to_burn_gdf = labels_to_burn_gdf.loc[
                labels_to_burn_gdf["burn_value"] != 0
            ].copy()

        elif len(classes) == 2:
            # There is no column with label names, but there are only 2 classes
            # (background + subject), so no problem...
            logger.info(
                f"Column ({labelname_column}) not found, so use all polygons in "
                f"{label_info.polygons_path.name}"
            )
            labels_to_burn_gdf.loc[:, "burn_value"] = 1

        else:
            # There is no column with label names, but more than two classes, so stop.
            validation_errors.append(
                f"Column {labelname_column} is mandatory in labeldata if multiple "
                f"classes specified: {classes}"
            )

        # Filter away None and empty geometries... they cannot be burned
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "GeoSeries.notna", UserWarning)
            labels_to_burn_gdf = labels_to_burn_gdf[
                ~labels_to_burn_gdf.geometry.is_empty
                & labels_to_burn_gdf.geometry.notna()
            ].copy()

        # Make sure all label polygons are valid
        labels_to_burn_gdf.loc[:, "is_valid_reason"] = vector_util.is_valid_reason(
            labels_to_burn_gdf.geometry
        )
        invalid_gdf = labels_to_burn_gdf.query("is_valid_reason != 'Valid Geometry'")
        for invalid in invalid_gdf.itertuples():
            validation_errors.append(
                f"Invalid geometry in {Path(invalid.path).name}: "
                f"{invalid.is_valid_reason}"
            )

        assert isinstance(labels_to_burn_gdf, gpd.GeoDataFrame)
        labeldata_result.append((labellocations_gdf, labels_to_burn_gdf))

    # Check that there is at least one train location and one validation location.
    if not train_locations_found:
        validation_errors.append(
            "No labellocations with traindata_type == 'train' found in any file! At "
            "least one needed"
        )
    if not validation_locations_found:
        validation_errors.append(
            "No labellocations with traindata_type == 'validation' found in any file! "
            "At least one needed, but ~10% of number of 'train' locations recommended."
        )

    # If errors found, raise
    if len(validation_errors) > 0:
        raise ValidationError(
            f"Errors found in label data: {len(validation_errors)}", validation_errors
        )

    return labeldata_result


def create_tmp_dir(
    parent_dir: Path, dir_name: str, remove_existing: bool = False
) -> Path:
    """Helper function to create a 'TMP' dir based on a directory name.

    The temp dir will be named like this:
        parent_dir / <dir_name>_TMP_<sequence>

    Use: if you want to write data to a directory in a "transactional way",
    it is the safest to write to a temp dir first, and then rename it to the
    final name. This way, if a hard crash occurs while writing the data, it
    is clear that the directory wasn't ready. Additionally, in case of a hard
    crash, file locks can remain which makes it impossible to remove a
    directory for a while.

    Args:
        parent_dir (Path): The dir to create the temp dir in.
        dir_name (str): The name of the dir to base the temp dir on.
        remove_existing (bool, optional): If True, existing TMP directories
            will be removed if possible. Defaults to False.

    Raises:
        Exception: [description]

    Returns:
        Path: [description]
    """
    # The final dir should not exists yet!
    final_dir = parent_dir / dir_name
    if final_dir.exists():
        raise Exception(
            f"It is not supported to create a TMP dir for an existing dir: {final_dir}"
        )

    # Try to delete all existing TMP dir's
    if remove_existing is True:
        existing_tmp_dirs = parent_dir.glob(f"{dir_name}_TMP_*")
        for existing_tmp_dir in existing_tmp_dirs:
            try:
                shutil.rmtree(existing_tmp_dir)
            except Exception:
                tmp_dir = None

    # Create the TMP dir
    tmp_dir = None
    for i in range(100):
        tmp_dir = parent_dir / f"{dir_name}_TMP_{i:02d}"

        if tmp_dir.exists():
            # If it (still) exists try next sequence
            tmp_dir = None
            continue
        else:
            # If it doesn't exist, try to create it
            try:
                tmp_dir.mkdir(parents=True)
                break
            except Exception:
                # If it fails to create, try next sequence
                tmp_dir = None
                continue

    # If no output tmp dir could be found/created... stop...
    if tmp_dir is None:
        raise Exception(
            f"Error creating/replacing TMP dir for {dir_name} in {parent_dir}"
        )

    return tmp_dir


def _create_mask(
    input_image_filepath: Path,
    output_mask_filepath: Path,
    labels_to_burn_gdf: gpd.GeoDataFrame,
    nb_classes: int = 1,
    output_imagecopy_filepath: Optional[Path] = None,
    minimum_pct_labeled: float = 0.0,
    force: bool = False,
) -> Optional[bool]:
    # If file exists already and force is False... stop.
    if force is False and output_mask_filepath.exists():
        logger.debug(
            f"Output file exist, and force is False, return: {output_mask_filepath}"
        )
        return None

    # Create a mask corresponding with the image file
    # First read the properties of the input image to copy them for the output
    logger.debug(f"Create mask to {output_mask_filepath}")
    with rio.open(input_image_filepath) as image_ds:
        image_input_profile = image_ds.profile
        image_transform_affine = image_ds.transform
        bounds = image_ds.bounds

    # Prepare the file profile for the mask depending on output type
    output_ext_lower = output_mask_filepath.suffix.lower()
    if output_ext_lower == ".tif":
        image_output_profile = rio_profiles.DefaultGTiffProfile(
            count=1, transform=image_transform_affine, crs=image_input_profile["crs"]
        )
    if output_ext_lower == ".png":
        image_output_profile = rio_profiles.Profile(driver="PNG", count=1)
    else:
        raise Exception(
            f"Unsupported mask suffix (should be lossless format!): {output_ext_lower}"
        )
    image_output_profile.update(
        width=image_input_profile["width"],
        height=image_input_profile["height"],
        dtype=rio.uint8,
    )

    # Filter the vectors that intersect the image bounds
    labels_to_burn_gdf = labels_to_burn_gdf.loc[
        labels_to_burn_gdf.intersects(
            sh_geom.box(bounds.left, bounds.bottom, bounds.right, bounds.top)
        )
    ].copy()

    # Burn the vectors in a mask
    burn_shapes = [
        (geom, value)
        for geom, value in zip(
            labels_to_burn_gdf.geometry, labels_to_burn_gdf.burn_value
        )
        if geom is not None and geom.is_empty is False
    ]
    if len(burn_shapes) > 0:
        try:
            mask_arr = rio_features.rasterize(
                shapes=burn_shapes,
                transform=image_transform_affine,
                dtype=rio.uint8,
                fill=0,
                out_shape=(
                    image_output_profile["width"],
                    image_output_profile["height"],
                ),
            )

        except Exception as ex:
            raise Exception(f"Error creating mask for {image_transform_affine}") from ex
    else:
        mask_arr = np.zeros(
            shape=(
                image_output_profile["width"],
                image_output_profile["height"],
            ),
            dtype=rio.uint8,
        )

    # Check if the mask meets the requirements to be written...
    if minimum_pct_labeled > 0:
        nb_pixels = np.size(mask_arr, 0) * np.size(mask_arr, 1)
        nb_pixels_data = nb_pixels - np.sum(
            mask_arr == 0
        )  # np.count_nonnan(image == NoData_value)
        logger.debug(
            f"nb_pixels: {nb_pixels}, nb_pixels_data: {nb_pixels_data}, "
            f"pct data: {nb_pixels_data / nb_pixels}"
        )

        if nb_pixels_data / nb_pixels < minimum_pct_labeled:
            return False

    # Write the labeled mask as .png (so without transform/crs info)
    im = Image.fromarray(mask_arr)
    im.save(output_mask_filepath)

    return True



"""
Helper functions for all tests.
"""

import logging
import tempfile
from pathlib import Path
from typing import ClassVar, Optional

import geopandas as gpd
from shapely import geometry as sh_geom

sampleprojects_dir = Path(__file__).resolve().parent.parent / "sample_projects"


class SampleProjectFootball:
    project_dir = sampleprojects_dir / "footballfields"
    predict_config_path = project_dir / "footballfields_BEFL-2019_test.ini"
    train_config_path = project_dir / "footballfields_train_test.ini"


class SampleProjectTemplate:
    project_dir = sampleprojects_dir / "project_template"


class TestData:
    dir = Path(__file__).resolve().parent / "data"

    classes: ClassVar = {
        "background": {
            "labelnames": ["ignore_for_train", "background"],
            "weight": 1,
            "burn_value": 0,
        },
        "test_classname1": {"labelnames": ["testlabel1"], "weight": 1, "burn_value": 1},
        "test_classname2": {"labelnames": ["testlabel2"], "weight": 1, "burn_value": 1},
    }
    image_pixel_x_size = 0.25
    image_pixel_y_size = 0.25
    image_pixel_width = 512
    image_pixel_height = 512
    image_crs_width = image_pixel_width * image_pixel_x_size
    image_crs_height = image_pixel_height * image_pixel_y_size
    crs_xmin = 150000
    crs_ymin = 150000
    crs = "EPSG:31370"
    location = sh_geom.box(
        crs_xmin,
        crs_ymin,
        crs_xmin + (image_pixel_width * image_pixel_x_size),
        crs_ymin + (image_pixel_height * image_pixel_y_size),
    )
    location_invalid = sh_geom.Polygon(
        [
            (crs_xmin, crs_ymin),
            (crs_xmin + image_crs_width, crs_ymin),
            (crs_xmin + image_crs_width, crs_ymin + image_crs_height),
            (crs_xmin, crs_ymin + image_crs_height),
            (
                crs_xmin + image_pixel_x_size,
                crs_ymin + image_crs_height + image_pixel_y_size,
            ),
            (crs_xmin, crs_ymin),
        ]
    )
    polygon = location
    polygon_invalid = location_invalid
    locations_gdf = gpd.GeoDataFrame(
        {
            "geometry": [location, location, location, location],
            "traindata_type": ["train", "validation", "test", "todo"],
            "path": "/tmp/locations.gdf",
        },
        crs="epsg:31370",
    )
    polygons_gdf = gpd.GeoDataFrame(
        {
            "geometry": [polygon, polygon],
            "classname": ["testlabel1", "testlabel2"],
            "path": "/tmp/polygons.gdf",
        },
        crs="epsg:31370",
    )


def create_tempdir(base_dirname: str, parent_dir: Optional[Path] = None) -> Path:
    # Parent
    if parent_dir is None:
        parent_dir = Path(tempfile.gettempdir())

    for i in range(1, 999999):
        try:
            tempdir = parent_dir / f"{base_dirname}_{i:06d}"
            tempdir.mkdir(parents=True)
            return Path(tempdir)
        except FileExistsError:
            continue

    raise Exception(
        "Wasn't able to create a temporary dir with basedir: "
        f"{parent_dir / base_dirname}"
    )


def init_test_for_debug(test_module_name: str) -> Path:
    # Init logging
    logging.basicConfig(
        format="%(asctime)s.%(msecs)03d|%(levelname)s|%(name)s|%(message)s",
        datefmt="%H:%M:%S",
        level=logging.INFO,
    )

    # Prepare tmpdir
    tmp_basedir = Path(tempfile.gettempdir()) / test_module_name
    tmpdir = create_tempdir(parent_dir=tmp_basedir, base_dirname="debugrun")

    """
    if tmpdir.exists():
        shutil.rmtree(tmpdir)
    tmpdir.mkdir(parents=True, exist_ok=True)
    """

    return tmpdir



import logging
import shutil
from pathlib import Path

import pytest

from orthoseg.helpers import config_helper as conf
from orthoseg.lib import cleanup
from tests import test_helper


def create_projects_dir(tmp_path: Path) -> Path:
    testproject_dir = tmp_path / "orthoseg_test_cleanup"
    project_dir = testproject_dir / "footballfields"

    shutil.rmtree(path=testproject_dir, ignore_errors=True)
    project_dir.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(
        src=test_helper.sampleprojects_dir / "project_template/projectfile.ini",
        dst=project_dir / "footballfields.ini",
    )
    shutil.copyfile(
        src=test_helper.sampleprojects_dir / "imagelayers.ini",
        dst=testproject_dir / "imagelayers.ini",
    )
    return project_dir


def create_model_files(dir: Path) -> list[tuple[Path, int]]:
    """
    Creates test model files and returns information about them.

    Args:
        dir (Path): the directory where the files should be created

    Returns:
        list[tuple[Path, int]]: the list of files created. Each item of the list is a
            tuple with:
              - the Path to the file
              - the minimum value of versions_to_retain that should lead to this file
                being retained when cleaning up the models.
    """
    models = {
        "footballfields_01": 4,
        "footballfields_02": 3,
        "footballfields_03": 2,
        "footballfields_04": 1,
    }
    model_files = [
        "0.44293_0.hdf5",
        "hyperparams.json",
        "log.csv",
        "model.json",
        "report.pdf",
    ]

    files = []
    dir.mkdir(parents=True, exist_ok=True)
    for model in models:
        for file_type in model_files:
            filepath = dir / f"{model}_{file_type}"
            filepath.touch()
            files.append((filepath, models[model]))

    return files


def create_training_dirs(dir: Path) -> list[tuple[Path, int]]:
    """
    Creates test training directories and returns information about them.

    Args:
        dir (Path): the directory where the directories should be created

    Returns:
        list[tuple[Path, int]]: the list of directories created. Each item of the list
            is a tuple with:
              - the Path to the directory
              - the minimum value of versions_to_retain that should lead to this file
                being retained when cleaning up the training directories.
    """
    # Directory paths with the minimum number of versions to retain to keep them
    dirs = [(dir / "01", 4), (dir / "02", 3), (dir / "03", 2), (dir / "04", 1)]

    # Create directories
    dir.mkdir(parents=True, exist_ok=True)
    for training_dir, _ in dirs:
        training_dir.mkdir(parents=True, exist_ok=True)

    return dirs


def create_prediction_files(dir: Path, imagelayer: str) -> list[tuple[Path, int]]:
    """
    Creates test prediction files and returns information about them.

    Args:
        dir (Path): the directory where the directories should be created

    Returns:
        list[tuple[Path, int]]: the list of files created. Each item of the list
            is a tuple with:
              - the Path to the file
              - the minimum value of versions_to_retain that should lead to this file
                being retained when cleaning up the prediction files.
    """
    # File paths with the minimum number of versions to retain to keep them
    files = [
        (dir / f"footballfields_01_201_{imagelayer}.gpkg", 4),
        (dir / f"footballfields_02_201_{imagelayer}.gpkg", 3),
        (dir / f"footballfields_03_201_{imagelayer}.gpkg", 2),
        (dir / f"footballfields_04_201_{imagelayer}.gpkg", 1),
        (dir / f"footballfields_01_201_{imagelayer}_dissolve.gpkg", 4),
        (dir / f"footballfields_02_201_{imagelayer}_dissolve.gpkg", 3),
        (dir / f"footballfields_03_201_{imagelayer}_dissolve.gpkg", 2),
        (dir / f"footballfields_04_201_{imagelayer}_dissolve.gpkg", 1),
    ]

    # Create the files
    dir.mkdir(parents=True, exist_ok=True)
    for file, _ in files:
        file.touch()

    return files


def load_project_config(path: Path):
    conf.read_orthoseg_config(
        config_path=path / "footballfields.ini",
        overrules=[
            "general.segment_subject=footballfields",
            "predict.image_layer=BEFL-2019",
        ],
    )


# Test parameters
# path : path of the filename that must be checked
# min_versions_to_retain_to_keep : if this version is less then versions_to_retain
#   the file (path) must still exist in the directory


def test_cleanup_non_existing_dir(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.INFO)
    caplog.clear()
    path = Path("not_existing")
    cleanup.clean_models(model_dir=path, versions_to_retain=3, simulate=True)
    assert f"Directory {path.name} doesn't exist" in caplog.text

    caplog.clear()
    cleanup.clean_training_data_directories(
        training_dir=path, versions_to_retain=3, simulate=True
    )
    assert f"Directory {path.name} doesn't exist" in caplog.text

    caplog.clear()
    path = Path("not_existing/not_existing")
    cleanup.clean_predictions(
        output_vector_dir=path, versions_to_retain=3, simulate=True
    )
    assert f"Directory {path.name} doesn't exist" in caplog.text


@pytest.mark.parametrize("simulate", [False, True])
@pytest.mark.parametrize("versions_to_retain", [4, 2, 1, 0, -1])
def test_cleanup_models(
    tmp_path: Path,
    simulate: bool,
    versions_to_retain: int,
):
    # Create test project
    project_dir = create_projects_dir(tmp_path=tmp_path)

    # Creating dummy files
    models_dir = project_dir / "models"
    model_files = create_model_files(dir=models_dir)

    # Load project config to init some vars.
    load_project_config(path=project_dir)

    # Cleanup
    files_removed = cleanup.clean_models(
        model_dir=conf.dirs.getpath("model_dir"),
        versions_to_retain=versions_to_retain,
        simulate=simulate,
    )

    # Asserts
    for path, min_version_to_retain in model_files:
        if versions_to_retain < 0 or min_version_to_retain <= versions_to_retain:
            # The file should be retained
            assert path not in files_removed
            assert path.exists()
        else:
            # The file should be cleaned up
            assert path in files_removed
            if simulate:
                assert path.exists()
            else:
                assert not path.exists()


@pytest.mark.parametrize("simulate", [False, True])
@pytest.mark.parametrize("versions_to_retain", [4, 2, 1, 0 - 1])
def test_cleanup_training(
    tmp_path: Path,
    simulate: bool,
    versions_to_retain: int,
):
    # Create test project
    project_dir = create_projects_dir(tmp_path=tmp_path)

    # Creating dummy files
    training_dir = project_dir / "training"
    dirs = create_training_dirs(dir=training_dir)

    # Load project config to init some vars.
    load_project_config(path=project_dir)

    # Cleanup
    cleanedup_trainingdata_dirs = cleanup.clean_training_data_directories(
        training_dir=conf.dirs.getpath("training_dir"),
        versions_to_retain=versions_to_retain,
        simulate=simulate,
    )

    # Asserts
    for dir, min_version_to_retain in dirs:
        if versions_to_retain < 0 or min_version_to_retain <= versions_to_retain:
            # Directory should be kept
            assert dir not in cleanedup_trainingdata_dirs
            assert dir.exists()
        else:
            # Directory should be removed
            assert dir in cleanedup_trainingdata_dirs
            if simulate:
                assert dir.exists()
            else:
                assert not dir.exists()


@pytest.mark.parametrize("simulate", [False, True])
@pytest.mark.parametrize("versions_to_retain", [4, 2, 1, 0 - 1])
def test_cleanup_predictions(
    tmp_path: Path,
    simulate: bool,
    versions_to_retain: int,
):
    # Create test project
    project_dir = create_projects_dir(tmp_path=tmp_path)

    # Creating dummy files
    imagelayer = "BEFL-2019"
    output_vector_dir = project_dir / "output_vector" / imagelayer
    files = create_prediction_files(dir=output_vector_dir, imagelayer=imagelayer)

    # Load project config to init some vars.
    load_project_config(path=project_dir)

    # Cleanup
    cleanedup_predictions = cleanup.clean_predictions(
        output_vector_dir=conf.dirs.getpath("output_vector_dir"),
        versions_to_retain=versions_to_retain,
        simulate=simulate,
    )

    # Asserts
    for path, min_version_to_retain in files:
        if versions_to_retain < 0 or min_version_to_retain <= versions_to_retain:
            # File should be kept
            assert path not in cleanedup_predictions
            if not simulate:
                assert path.exists()
        else:
            # File should be removed
            assert path in cleanedup_predictions
            if simulate:
                assert path.exists()
            else:
                assert not path.exists()


@pytest.mark.parametrize("simulate", [False])
@pytest.mark.parametrize(
    "versions_to_retain, removed_model_files, removed_training_dirs, "
    "removed_prediction_files",
    [
        (4, 0, 0, 0),
        (2, 2 * 5, 2, 8),
        (1, 3 * 5, 3, 12),
        (0, 4 * 5, 4, 16),
        (-1, 0, 0, 0),
    ],
)
def test_cleanup_project_dir(
    tmp_path: Path,
    simulate: bool,
    versions_to_retain: int,
    removed_model_files: int,
    removed_training_dirs: int,
    removed_prediction_files: int,
):
    # Create test project
    project_dir = create_projects_dir(tmp_path=tmp_path)

    # Creating dummy files
    model_dir = project_dir / "models"
    create_model_files(dir=model_dir)
    training_dir = project_dir / "training"
    create_training_dirs(dir=training_dir)
    imagelayer = "BEFL-2019"
    output_vector_dir = project_dir / "output_vector" / imagelayer
    create_prediction_files(dir=output_vector_dir, imagelayer=imagelayer)
    imagelayer = "BEFL-2020"
    output_vector_dir = project_dir / "output_vector" / imagelayer
    create_prediction_files(dir=output_vector_dir, imagelayer=imagelayer)

    # Load project config to init some vars.
    load_project_config(path=project_dir)

    removed = cleanup.clean_project_dir(
        model_dir=conf.dirs.getpath("model_dir"),
        model_versions_to_retain=versions_to_retain,
        training_dir=conf.dirs.getpath("training_dir"),
        training_versions_to_retain=versions_to_retain,
        output_vector_dir=conf.dirs.getpath("output_vector_dir"),
        prediction_versions_to_retain=versions_to_retain,
        simulate=simulate,
    )

    # Asserts
    assert len(removed["models"]) == removed_model_files
    assert len(removed["training_dirs"]) == removed_training_dirs
    assert len(removed["predictions"]) == removed_prediction_files



"""
Tests for functionalities in orthoseg.lib.postprocess_predictions.
"""

import os

import geofileops as gfo
import geopandas as gpd
import pandas as pd
from shapely import geometry as sh_geom

# Make hdf5 version warning non-blocking
os.environ["HDF5_DISABLE_VERSION_CHECK"] = "1"

from orthoseg.helpers import vectorfile_helper


def test_reclassify_neighbours(tmp_path):
    # Prepare test data
    reclassify_column = "class_name"
    query = "area <= 20"
    columns = ["test_id", "descr", "expected_result", "geometry", reclassify_column]
    testdata = [
        ["None1", "Geom None", "disappears", None, "class_1"],
        ["None2", "Geom None", "disappears", None, "class_1"],
        ["None3", "to be removed", "disappears", None, "class_1"],
        ["Empty1", "Polygon EMPTY", "disappears", sh_geom.Polygon(), "class_1"],
        [
            "poly1",
            "large polygon (25m²)",
            "smaller polygons are added to it",
            sh_geom.Polygon(shell=[(5, 5), (10, 5), (10, 10), (5, 10), (5, 5)]),
            "class_1_large",
        ],
        [
            "poly2",
            "medium (5m²), neighbour of polygon1, onborder",
            "is onborder, so isn't merged with other things",
            sh_geom.Polygon(shell=[(0, 5), (0, 6), (5, 6), (5, 5), (0, 5)]),
            "class_2_medium",
        ],
        [
            "poly3",
            "neighbour of polygon1, small (1m²), not onborder",
            "is merged with poly1",
            sh_geom.Polygon(shell=[(10, 5), (11, 5), (11, 6), (10, 6), (10, 5)]),
            "class_3_small",
        ],
        [
            "poly4",
            "neighbour of poly3, small (1m²), not onborder",
            "is merged with poly1 and poly3",
            sh_geom.Polygon(shell=[(11, 5), (12, 5), (12, 6), (11, 6), (11, 5)]),
            "class_4_small",
        ],
        [
            "poly5",
            "medium polygon (5m²), larger than neighbour poly6, not onborder",
            "is merged with poly6 but retains class",
            sh_geom.Polygon(shell=[(5, 15), (10, 15), (10, 16), (5, 16), (5, 15)]),
            "class_5_medium",
        ],
        [
            "poly6",
            "small (1m²), smaller than neighbour poly5, not onborder",
            "is merged with poly6 and gets its class",
            sh_geom.Polygon(shell=[(4, 15), (5, 15), (5, 16), (4, 16), (4, 15)]),
            "class_6_small",
        ],
    ]
    df = pd.DataFrame(testdata, columns=columns)
    testdata_gdf = gpd.GeoDataFrame(df, geometry="geometry")
    # Remove a row to test if the index is properly maintained
    testdata_gdf = testdata_gdf.drop([2], axis=0)
    testdata_path = tmp_path / "testdata.gpkg"
    gfo.to_file(testdata_gdf, testdata_path)

    # Test
    result_path = tmp_path / "result.gpkg"
    vectorfile_helper.reclassify_neighbours(
        input_path=testdata_path,
        reclassify_column=reclassify_column,
        query=query,
        output_path=result_path,
    )
    result_gdf = gfo.read_file(result_path)
    assert result_gdf is not None
    assert len(result_gdf) == 2
    assert reclassify_column in result_gdf.columns
    assert "area" in result_gdf.columns
    assert "nbcoords" in result_gdf.columns



import re

import pytest

from orthoseg.helpers import config_helper as conf
from orthoseg.lib.prepare_traindatasets import LabelInfo
from tests.test_helper import SampleProjectFootball, TestData


@pytest.mark.parametrize(
    "overrules, exp_error, message",
    [
        (
            ["general.segment_subject=MUST_OVERRIDE"],
            ValueError,
            "Projectconfig parameter general.segment_subject needs to be overruled",
        ),
        (
            ["general.segment_subject=Football_field"],
            ValueError,
            "should not contain any of the following chars",
        ),
    ],
)
def test_read_orthoseg_config_error_segment_subject(overrules, exp_error, message):
    with pytest.raises(
        exp_error,
        match=message,
    ):
        conf.read_orthoseg_config(
            SampleProjectFootball.predict_config_path, overrules=overrules
        )


def test_read_orthoseg_config_image_layers():
    # Load project config to init some vars.
    conf.read_orthoseg_config(SampleProjectFootball.predict_config_path)

    layer = conf.image_layers.get("BEFL-2019")
    assert layer is not None


@pytest.mark.parametrize(
    "overrules, expected_image_layer",
    [
        (None, "BEFL-2019"),
        ([], "BEFL-2019"),
        (["predict.image_layer=BEFL-2020"], "BEFL-2020"),
    ],
)
def test_read_orthoseg_config_predict_overrules(overrules, expected_image_layer):
    # Load project config to test overrules.
    kwargs = {}
    if overrules is not None:
        kwargs["overrules"] = overrules
    conf.read_orthoseg_config(SampleProjectFootball.predict_config_path, **kwargs)

    image_layer = conf.predict.get("image_layer")
    assert image_layer == expected_image_layer


@pytest.mark.parametrize(
    "overrules",
    [["predictimage_layer=BEFL-2020"], ["predict.image_layerBEFL-2020"]],
)
def test_read_orthoseg_config_predict_overrules_invalid(overrules):
    """Test with invalid overrules: one without '=' and one without '.'."""
    # Load project config to test overrules.
    with pytest.raises(ValueError, match="invalid config overrule found"):
        conf.read_orthoseg_config(
            SampleProjectFootball.predict_config_path, overrules=overrules
        )


@pytest.mark.parametrize(
    "overrules, expected_image_layer",
    [
        (None, None),
        ([], None),
        (["train.image_layer=BEFL-2020"], "BEFL-2020"),
    ],
)
def test_read_orthoseg_config_train_overrules(overrules, expected_image_layer):
    # Load project config to test overrules.
    kwargs = {}
    if overrules is not None:
        kwargs["overrules"] = overrules
    conf.read_orthoseg_config(SampleProjectFootball.train_config_path, **kwargs)

    image_layer = conf.train.get("image_layer")
    if expected_image_layer is None:
        assert image_layer is None
    else:
        assert image_layer == expected_image_layer


def test_prepare_train_label_infos():
    labelpolygons_pattern = TestData.dir / "footballfields_{image_layer}_data.gpkg"
    labellocations_pattern = (
        TestData.dir / "footballfields_{image_layer}_locations.gpkg"
    )
    image_layers = {"BEFL-2019": {}, "BEFL-2020": {}, "BEFL-2021": {}, "BEFL-2022": {}}
    label_datasources = {
        "label_ds1": {
            "locations_path": str(
                TestData.dir / "footballfields_BEFL-2019_locations.gpkg"
            ),
            "polygons_path": str(TestData.dir / "footballfields_BEFL-2019_data.gpkg"),
            "pixel_x_size": 1,
            "pixel_y_size": 2,
            "image_layer": "BEFL-2021",
        },
        "label_ds2": {
            "locations_path": str(
                TestData.dir / "footballfields_BEFL-2022_locations.gpkg"
            ),
            "data_path": str(TestData.dir / "footballfields_BEFL-2022_data.gpkg"),
            "pixel_x_size": 5,
            "pixel_y_size": 6,
            "image_layer": "BEFL-2022",
        },
    }

    label_infos_result = conf._prepare_train_label_infos(
        labelpolygons_pattern,
        labellocations_pattern,
        label_datasources=label_datasources,
        image_layers=image_layers,
    )
    assert len(label_infos_result) == 3
    for result in label_infos_result:
        if result.image_layer == "BEFL-2021":
            assert result.pixel_x_size == 1
            assert result.pixel_y_size == 2
        elif result.image_layer == "BEFL-2022":
            assert result.pixel_x_size == 5
            assert result.pixel_y_size == 6
        else:
            assert result.pixel_x_size is None
            assert result.pixel_y_size is None


def test_prepare_train_label_infos_invalid_layer():
    labeldata_pattern = TestData.dir / "footballfields_{image_layer}_data.gpkg"
    labellocation_pattern = TestData.dir / "footballfields_{image_layer}_locations.gpkg"
    image_layers = {"BEFL-2019": {}}

    with pytest.raises(
        ValueError, match=re.escape("invalid image_layer in LabelInfo(image_layer=")
    ):
        _ = conf._prepare_train_label_infos(
            labeldata_pattern,
            labellocation_pattern,
            label_datasources={},
            image_layers=image_layers,
        )


def test_search_label_files():
    labelpolygons_pattern = TestData.dir / "footballfields_{image_layer}_data.gpkg"
    labellocation_pattern = TestData.dir / "footballfields_{image_layer}_locations.gpkg"
    results = conf._search_label_files(labelpolygons_pattern, labellocation_pattern)

    assert len(results) == 2
    for result in results:
        assert isinstance(result, LabelInfo)


def test_search_label_files_invalid_dir():
    invalid_dir = TestData.dir / "unexisting"
    labelpolygons_pattern = TestData.dir / "footballfields_{image_layer}_data.gpkg"
    labellocation_pattern = invalid_dir / "footballfields_{image_layer}_locations.gpkg"

    with pytest.raises(ValueError, match="Label dir doesn't exist"):
        _ = conf._search_label_files(labelpolygons_pattern, labellocation_pattern)

    labelpolygons_pattern = invalid_dir / "footballfields_{image_layer}_data.gpkg"
    labellocation_pattern = TestData.dir / "footballfields_{image_layer}_locations.gpkg"

    with pytest.raises(ValueError, match="Label dir doesn't exist"):
        _ = conf._search_label_files(labelpolygons_pattern, labellocation_pattern)


def test_unformat():
    result = conf._unformat(
        "footballfields_BEFL-2018_data.gpkg",
        pattern="footballfields_{image_layer}_data.gpkg",
    )
    assert result == {"image_layer": "BEFL-2018"}


def test_unformat_error():
    with pytest.raises(
        ValueError, match="pattern fields_{image_layer}_data.gpkg not found"
    ):
        _ = conf._unformat(
            "fields_BEFL-2018_polygons.gpkg",
            pattern="fields_{image_layer}_data.gpkg",
        )



import os
import shutil
from datetime import datetime

import pytest

import orthoseg
from orthoseg.helpers import config_helper as conf
from orthoseg.validate import _validate_args
from tests import test_helper
from tests.test_helper import SampleProjectFootball


def test_validate(tmp_path):
    # Copy footballfields sample project for this test
    testprojects_dir = tmp_path / "sample_projects"
    footballfields_dir = testprojects_dir / "footballfields"
    shutil.copytree(test_helper.sampleprojects_dir, testprojects_dir)

    # Load project config to init some vars.
    config_path = footballfields_dir / SampleProjectFootball.train_config_path.name
    conf.read_orthoseg_config(config_path)

    # Init + cleanup result dirs
    traindata_id_result = 2
    training_dir = conf.dirs.getpath("training_dir")
    training_id_dir = training_dir / f"{traindata_id_result:02d}"
    if training_id_dir.exists():
        shutil.rmtree(training_id_dir)
    model_dir = conf.dirs.getpath("model_dir")
    if model_dir.exists():
        modelfile_paths = model_dir.glob(f"footballfields_{traindata_id_result:02d}_*")
        for modelfile_path in modelfile_paths:
            modelfile_path.unlink()

    # Make sure the label files in version 01 are older than those in the label dir
    # so a new model will be trained
    label_01_path = training_dir / "01/footballfields_BEFL-2019_polygons.gpkg"
    timestamp_old = datetime(year=2020, month=1, day=1).timestamp()
    os.utime(label_01_path, (timestamp_old, timestamp_old))

    # Run validate
    orthoseg.validate(config_path=config_path)


@pytest.mark.parametrize(
    "args",
    [
        (
            [
                "--config",
                "X:/Monitoring/OrthoSeg/test/test.ini",
                "predict.image_layer=LT-2023",
            ]
        )
    ],
)
def test_validate_args(args):
    valid_args = _validate_args(args=args)

    assert valid_args is not None
    assert valid_args.config is not None
    assert valid_args.config_overrules is not None


def test_validate_error(tmp_path):
    # Create test project
    project_dir = tmp_path / "footballfields"
    project_dir.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(
        src=test_helper.SampleProjectFootball.project_dir
        / "footballfields_BEFL-2019.ini",
        dst=project_dir / "footballfields_BEFL-2019.ini",
    )
    shutil.copyfile(
        src=test_helper.sampleprojects_dir / "imagelayers.ini",
        dst=tmp_path / "imagelayers.ini",
    )

    with pytest.raises(
        Exception,
        match="ERROR while running validate for task footballfields_BEFL-2019",
    ):
        orthoseg.validate(
            config_path=project_dir / "footballfields_BEFL-2019.ini",
        )



"""
Tests for functionalities in orthoseg.lib.postprocess_predictions.
"""

import os
import shutil
from pathlib import Path

import geofileops as gfo
import geopandas as gpd
import pandas as pd
import pytest
import shapely

from orthoseg.lib import postprocess_predictions as postp
from tests import test_helper

# Make hdf5 version warning non-blocking
os.environ["HDF5_DISABLE_VERSION_CHECK"] = "1"

from orthoseg.lib import postprocess_predictions as post_pred
from tests.test_helper import TestData


def test_read_prediction_file():
    # Read + polygonize raster prediction file
    pred_raster_path = TestData.dir / "129568_185248_130592_186272_4096_4096_1_pred.tif"
    pred_raster_gdf = post_pred.read_prediction_file(
        pred_raster_path, border_pixels_to_ignore=128
    )
    # gfo.to_file(pred_raster_gdf, get_testdata_dir() / f"{pred_raster_path.stem}.gpkg")

    # Read the comparison file, that contains the result of the polygonize
    pred_comparison_path = TestData.dir / f"{pred_raster_path.stem}.gpkg"
    pred_comparison_gdf = gfo.read_file(pred_comparison_path)

    # Now compare they are the same
    assert pred_raster_gdf is not None
    assert len(pred_raster_gdf) == len(pred_comparison_gdf)


def test_clean_vectordata(tmpdir):
    temp_dir = Path(tmpdir)

    # Clean data
    input1_path = TestData.dir / "129568_184288_130592_185312_4096_4096_1_pred.gpkg"
    input2_path = TestData.dir / "129568_185248_130592_186272_4096_4096_1_pred.gpkg"
    input1_gdf = gfo.read_file(input1_path)
    input2_gdf = gfo.read_file(input2_path)
    input_gdf = pd.concat([input1_gdf, input2_gdf])
    assert input1_gdf.crs == input_gdf.crs
    input_path = temp_dir / "vector_input.gpkg"
    gfo.to_file(input_gdf, input_path)
    output_path = temp_dir / input_path.name
    post_pred.postprocess_predictions(
        input_path=input_path,
        output_path=output_path,
        dissolve=True,
        dissolve_tiles_path=None,
        force=True,
    )

    # Read result and check
    geoms_simpl_filepath = (
        output_path.parent / f"{output_path.stem}_dissolve{output_path.suffix}"
    )
    result_gdf = gfo.read_file(geoms_simpl_filepath)

    assert len(result_gdf) == 616


def create_projects_dir(tmp_path: Path) -> Path:
    testproject_dir = tmp_path / "orthoseg_test_postprocess"
    project_dir = testproject_dir / "footballfields"

    shutil.rmtree(path=testproject_dir, ignore_errors=True)
    project_dir.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(
        src=test_helper.sampleprojects_dir / "project_template/projectfile.ini",
        dst=project_dir / "footballfields.ini",
    )
    shutil.copyfile(
        src=test_helper.sampleprojects_dir / "imagelayers.ini",
        dst=testproject_dir / "imagelayers.ini",
    )
    return project_dir


def create_prediction_file(output_vector_dir: Path) -> Path:
    imagelayer = "BEFL-2019"
    prediction_dir = output_vector_dir / imagelayer
    prediction_dir.mkdir(parents=True, exist_ok=True)
    output_vector_path = prediction_dir / f"footballfields_01_201_{imagelayer}.gpkg"

    footballfields = {
        "classname": [
            "footballfields",
            "footballfields",
            "footballfields",
            "footballfields",
            "footballfields",
        ],
        "geometry": [
            shapely.wkt.loads(
                "MULTIPOLYGON (((175056.375 176371.125, 175056.375 176370.875, 175056.625 176370.875, 175056.625 176371.125, 175056.375 176371.125)))"  # noqa: E501
            ),
            shapely.wkt.loads(
                "MULTIPOLYGON (((175054.875 176370.875, 175054.875 176370.625, 175055.375 176370.625, 175055.375 176370.875, 175054.875 176370.875)))"  # noqa: E501
            ),
            shapely.wkt.loads(
                "MULTIPOLYGON (((175047.125 176370.875, 175047.125 176370.125, 175047.375 176370.625, 175047.125 176370.875)))"  # noqa: E501
            ),
            shapely.wkt.loads(
                "MULTIPOLYGON (((175046.625 176370.125, 175046.625 176369.875, 175046.875 176369.875, 175046.875 176370.125, 175046.625 176370.125)))"  # noqa: E501
            ),
            shapely.wkt.loads(
                "MULTIPOLYGON (((175054.625 176369.625, 175054.625 176368.875, 175055.625 176369.375, 175054.625 176369.625)))"  # noqa: E501
            ),
        ],
    }
    footballfields_gdf = gpd.GeoDataFrame(footballfields, geometry="geometry")
    gfo.to_file(gdf=footballfields_gdf, path=output_vector_path)
    return output_vector_path


@pytest.mark.parametrize("keep_original_file", [False, True])
@pytest.mark.parametrize("keep_intermediary_files", [False, True])
def test_postprocess_predictions(
    tmp_path: Path,
    keep_original_file: bool,
    keep_intermediary_files: bool,
):
    # Create test project
    project_dir = create_projects_dir(tmp_path=tmp_path)

    # Creating dummy files
    output_vector_dir = project_dir / "output_vector"
    output_vector_path = create_prediction_file(output_vector_dir=output_vector_dir)
    output_orig_path = (
        output_vector_path.parent / f"{output_vector_path.stem}_orig.gpkg"
    )
    output_dissolve_path = (
        output_vector_path.parent / f"{output_vector_path.stem}_dissolve.gpkg"
    )
    output_reclass_path = (
        output_vector_path.parent / f"{output_vector_path.stem}_reclass.gpkg"
    )

    # Go!
    postp.postprocess_predictions(
        input_path=output_vector_path,
        output_path=output_vector_path,
        keep_original_file=keep_original_file,
        keep_intermediary_files=keep_intermediary_files,
        dissolve=True,
        reclassify_to_neighbour_query="(area < 5)",
        force=False,
    )

    # Check results
    if not keep_original_file and not keep_intermediary_files:
        assert len(os.listdir(output_vector_path.parent)) == 1
        assert output_vector_path.exists()
    if keep_original_file and not keep_intermediary_files:
        assert len(os.listdir(output_vector_path.parent)) == 2
        assert output_vector_path.exists()
        assert output_orig_path.exists()
    if not keep_original_file and keep_intermediary_files:
        assert len(os.listdir(output_vector_path.parent)) == 3
        assert output_vector_path.exists()
        assert output_dissolve_path.exists()
        assert output_reclass_path.exists()
    if keep_original_file and keep_intermediary_files:
        assert len(os.listdir(output_vector_path.parent)) == 4
        assert output_vector_path.exists()
        assert output_orig_path.exists()
        assert output_dissolve_path.exists()
        assert output_reclass_path.exists()



"""Module with specific helper functions to manage the configuration of orthoseg."""

import configparser
import json
import logging
import tempfile
from pathlib import Path
from typing import Optional

# Get a logger...
logger = logging.getLogger(__name__)

# Define the chars that cannot be used in codes that are use in filenames.
# Remark: '_' cannot be used because '_' is used as devider to parse filenames, and if
# it is used in codes as well the parsing becomes a lot more difficult.
illegal_chars_in_codes = ["_", ",", ".", "?", ":"]


def get_config_files(config_path: Path) -> list[Path]:
    """Get a list of all relevant config files.

    The list of files returned is of importance:
    - first the general "project_defaults.ini" file packaged in orthoseg
    - then all configuration files in property [general].extra_config_files_to_load of
      ``config_path``, in the order they are listed there
    - finally ``config_path`` itself

    The main principle is that the configuration files are ordered by importance.
    Configuration in the files later in the list will overrule configuration in the
    prior ones.

    Args:
        config_path (Path): base config file.

    Returns:
        list[Path]: all relevant config files.
    """
    # Init
    # First check input param
    config_path = config_path.expanduser()
    if not config_path.exists():
        raise ValueError(f"Config file specified does not exist: {config_path}")

    # Collect the config files to use. The "hardcoded" defaults should always
    # be loaded.
    script_dir = Path(__file__).resolve().parent.parent
    config_filepaths = [script_dir / "project_defaults.ini"]

    # Load possible extra config files from the project config file
    basic_projectconfig = configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation(),
        converters={"list": lambda x: [i.strip() for i in x.split(",")]},
        allow_no_value=True,
    )
    basic_projectconfig.read(config_path)

    default_basedir = config_path.parent
    extra_config_files_to_load = basic_projectconfig["general"].getlist(
        "extra_config_files_to_load"
    )
    if extra_config_files_to_load is not None:
        for config_file in extra_config_files_to_load:
            config_file_formatted = Path(
                config_file.format(
                    task_filepath=config_path, jobs_dir=config_path.parent
                )
            ).expanduser()
            if not config_file_formatted.is_absolute():
                config_file_formatted = (
                    default_basedir / config_file_formatted
                ).resolve()
            config_filepaths.append(Path(config_file_formatted))

    # Finally add the specific project config file...
    config_filepaths.append(config_path)

    return config_filepaths


def read_config_ext(config_paths: list[Path]) -> configparser.ConfigParser:
    """Read configuration with extended functionalities.

    Args:
        config_paths (list[Path]): the configuration files to load.

    Returns:
        configparser.ConfigParser: _description_
    """
    # Log config filepaths that don't exist...
    for config_filepath in config_paths:
        if not config_filepath.exists():
            logger.warning(f"config_filepath does not exist: {config_filepath}")

    # Now we are ready to read the entire configuration
    def parse_boolean_ext(input) -> Optional[bool]:
        if input is None:
            return None

        if input in ("True", "true", "1", 1):
            return True
        elif input in ("False", "false", "0", 0):
            return False
        else:
            return None

    def safe_math_eval(string):
        """Function to evaluate a mathematical expression safely."""
        if string is None:
            return None

        allowed_chars = "0123456789+-*(). /"
        for char in string:
            if char not in allowed_chars:
                raise ValueError("Error: Unsafe eval")

        return eval(string)

    def to_path(pathlike: str) -> Optional[Path]:
        if pathlike is None:
            return None
        else:
            if "{tempdir}" in pathlike:
                return Path(pathlike.format(tempdir=tempfile.gettempdir()))
            else:
                return Path(pathlike)

    config = configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation(),
        converters={
            "list": lambda x: [i.strip() for i in x.split(",")],
            "listint": lambda x: [int(i.strip()) for i in x.split(",")],
            "listfloat": lambda x: [float(i.strip()) for i in x.split(",")],
            "dict": lambda x: None if x is None else json.loads(x),
            "path": lambda x: to_path(x),
            "eval": lambda x: safe_math_eval(x),
            "boolean_ext": lambda x: parse_boolean_ext(x),
        },
        allow_no_value=True,
    )

    config.read(config_paths)
    return config


def as_dict(config: configparser.ConfigParser):
    """Converts a ConfigParser object into a dictionary.

    The resulting dictionary has sections as keys which point to a dict of the
    sections options as key => value pairs.
    """
    the_dict: dict = {}
    for section in config.sections():
        the_dict[section] = {}
        for key, val in config.items(section):
            the_dict[section][key] = val
    return the_dict



"""Module with functions to clean up old project data."""

import logging
import os
import shutil
from pathlib import Path

from orthoseg.model import model_helper
from orthoseg.util.data import AiDetectionInfo, aidetection_info

# Get a logger...
logger = logging.getLogger(__name__)


def clean_models(
    model_dir: Path, versions_to_retain: int, simulate: bool
) -> list[Path]:
    """Cleanup models.

    Args:
        model_dir (Path): Path to the directory with the models to be cleaned
        versions_to_retain (int): Number of versions to retain. If <0, all versions are
            retained.
        simulate (bool): Simulate cleanup, files are logged, no files are deleted

    Raises:
        Exception: ERROR while deleting file

    Returns:
        list[Path]: List of removed model files
    """
    # Check input
    if versions_to_retain < 0:
        return []
    if not model_dir.exists():
        logger.info(f"Directory {model_dir.name} doesn't exist")
        return []

    logger.info(f"clean_models with {model_dir=}, {versions_to_retain=}, {simulate=}")
    models_to_cleanup = []
    files_to_remove = []
    models = model_helper.get_models(model_dir=model_dir)
    traindata_id = [model["traindata_id"] for model in models]
    traindata_id.sort()
    traindata_id_to_cleanup = traindata_id[
        : len(traindata_id) - versions_to_retain
        if len(traindata_id) >= versions_to_retain
        else 0
    ]
    models_to_cleanup = [
        model["basefilename"]
        for model in models
        if model["traindata_id"] in traindata_id_to_cleanup
    ]

    # Find all files to remove and remove them
    for model in models_to_cleanup:
        file_pattern = f"{model}*.*"
        files = list(model_dir.glob(pattern=file_pattern))
        files_to_remove.extend(files)

        for file_to_remove in files:
            logger.info(f"remove prediction file ({simulate=}): {file_to_remove}")
            if simulate:
                continue

            try:
                file_to_remove.unlink()
            except OSError as ex:
                message = f"ERROR while deleting file {file_to_remove}"
                logger.exception(message)
                raise RuntimeError(message) from ex

    return files_to_remove


def clean_training_data_directories(
    training_dir: Path, versions_to_retain: int, simulate: bool
) -> list[Path]:
    """Cleanup training data directories.

    Args:
        training_dir (Path): Path to the directory with the training data
        versions_to_retain (int): Number of versions to retain. If <0, all versions are
            retained.
        simulate (bool): Simulate cleanup, directories are logged, no files are deleted

    Raises:
        Exception: ERROR while deleting directory

    Returns:
        list[Path]: List of training directories to be removed
    """
    # Check input
    if versions_to_retain < 0:
        return []
    if not training_dir.exists():
        logger.info(f"Directory {training_dir.name} doesn't exist")
        return []

    logger.info(
        f"clean_training_data_directories with {training_dir=}, {versions_to_retain=}, "
        f"{simulate=}"
    )
    dirnames = [dir for dir in os.listdir(training_dir) if dir.isnumeric()]
    dirnames.sort()
    dirnames_to_clean = []
    dirnames_to_clean = dirnames[
        : len(dirnames) - versions_to_retain
        if len(dirnames) >= versions_to_retain
        else 0
    ]

    dirs_to_remove = []
    for dirname in dirnames_to_clean:
        dir = training_dir / dirname
        dirs_to_remove.append(dir)
        logger.info(f"remove training dir ({simulate=}): {dir}")

        if simulate:
            continue

        try:
            shutil.rmtree(dir)
        except Exception as ex:
            message = f"ERROR deleting directory {dir}"
            logger.exception(message)
            raise RuntimeError(message) from ex

    return dirs_to_remove


def clean_predictions(
    output_vector_dir: Path, versions_to_retain: int, simulate: bool
) -> list[Path]:
    """Cleanup predictions.

    Args:
        output_vector_dir (Path): Path to the directory containing the vector
            predictions
        versions_to_retain (int): Number of versions to retain. If <0, all versions are
            retained.
        simulate (bool): Simulate cleanup, files are logged, no files are deleted

    Raises:
        Exception: ERROR while deleting file

    Returns:
        list[Path]: List of prediction files to be removed
    """
    # Check input
    if versions_to_retain < 0:
        return []
    if not output_vector_dir.exists():
        logger.info(f"Directory {output_vector_dir.name} doesn't exist")
        return []

    logger.info(
        f"clean_predictions with {output_vector_dir=}, {versions_to_retain=}, "
        f"{simulate=}"
    )
    predictions_to_cleanup: list[AiDetectionInfo] = []
    prediction_files_to_remove = []
    files = output_vector_dir.glob(pattern="*.*")
    try:
        ai_detection_infos = [aidetection_info(path=file) for file in files]
        postprocessing = [x.postprocessing for x in ai_detection_infos]
        postprocessing = list(dict.fromkeys(postprocessing))
        logger.info(f"{output_vector_dir=}, {versions_to_retain=}, {simulate=}")
        for p in postprocessing:
            traindata_versions = [
                ai_detection_info.traindata_version
                for ai_detection_info in ai_detection_infos
                if p == ai_detection_info.postprocessing
            ]
            traindata_versions.sort()
            traindata_versions_to_cleanup = traindata_versions[
                : len(traindata_versions) - versions_to_retain
                if len(traindata_versions) >= versions_to_retain
                else 0
            ]
            predictions_to_cleanup.extend(
                [
                    ai_detection_info
                    for ai_detection_info in ai_detection_infos
                    if ai_detection_info.traindata_version
                    in traindata_versions_to_cleanup
                    and ai_detection_info.postprocessing == p
                ]
            )
        for prediction in predictions_to_cleanup:
            prediction_path = prediction.path
            prediction_files_to_remove.append(prediction_path)
            logger.info(f"remove prediction file ({simulate=}): {prediction_path.name}")
            if simulate:
                continue

            try:
                prediction_path.unlink()
            except Exception as ex:
                message = f"ERROR while deleting file {prediction_path}"
                logger.exception(message)
                raise RuntimeError(message) from ex

    except Exception as ex:
        logger.info(f"{ex}")

    return prediction_files_to_remove


def clean_project_dir(
    model_dir: Path,
    model_versions_to_retain: int,
    training_dir: Path,
    training_versions_to_retain: int,
    output_vector_dir: Path,
    prediction_versions_to_retain: int,
    simulate: bool,
) -> dict:
    """Cleanup project directory.

    Args:
        model_dir (Path): Path to the directory with the models to be cleaned
        model_versions_to_retain (int): Model versions to retain. If <0, all models are
            retained.
        training_dir (Path): Path to the directory with the training data to be cleaned
        training_versions_to_retain (int): Training data versions to retain. If <0, all
            versions are retainded.
        output_vector_dir (Path): Path to the directory with the predictions to be
            cleaned
        prediction_versions_to_retain (int): Prediction versions to retain. If <0, all
            versions are retained.
        simulate (bool): Simulate cleanup, files are logged, no files are deleted

    Returns:
        dict: Dictionary with removed models, training directories and predictions
    """
    removed = {}

    removed["models"] = clean_models(
        model_dir=model_dir,
        versions_to_retain=model_versions_to_retain,
        simulate=simulate,
    )

    removed["training_dirs"] = clean_training_data_directories(
        training_dir=training_dir,
        versions_to_retain=training_versions_to_retain,
        simulate=simulate,
    )

    output_vector_parent_dir = output_vector_dir.parent
    removed["predictions"] = []
    if output_vector_parent_dir.exists():
        prediction_dirs = os.listdir(output_vector_parent_dir)
        for prediction_dir in prediction_dirs:
            removed["predictions"].extend(
                clean_predictions(
                    output_vector_dir=output_vector_parent_dir / prediction_dir,
                    versions_to_retain=prediction_versions_to_retain,
                    simulate=simulate,
                )
            )
    else:
        logger.info(f"Directory {output_vector_parent_dir.name} doesn't exist")

    return removed



"""Download the sample project."""

import argparse
import logging
import shlex
import sys
from pathlib import Path
from typing import Optional

import gdown

from orthoseg.util import git_downloader

# Get a logger...
logger = logging.getLogger(__name__)


def _parse_load_sampleprojects_argstr(argstr):
    args = shlex.split(argstr)
    _parse_load_sampleprojects_args(args)


def _parse_load_sampleprojects_args(args) -> dict:
    # Define supported arguments
    parser = argparse.ArgumentParser(add_help=False)

    help = (
        "The directory to create the sample_projects dir in. "
        "Eg. ~/orthoseg will create orthoseg/sample_projects in your home directory."
    )
    parser.add_argument("dest_dir", help=help)

    help = (
        "True to use the default certificate bundle as installed on your system. "
        "False disables certificate validation (NOT recommended!). In corporate "
        "networks using a proxy server it is often needed to specify a customized "
        "certificate bundle (.pem file) to avoid CERTIFICATE_VERIFY_FAILED errors. "
        "It is recommended to specify the path to a custum certificate bundle file "
        "using the REQUESTS_CA_BUNDLE environment variable, but it can also passed "
        "using this switch. Parameter defaults to True."
    )
    parser.add_argument("--ssl_verify", default=True, help=help)

    # Interprete arguments
    args = parser.parse_args(args)
    dest_dir = Path(args.dest_dir).expanduser() / "orthoseg"
    ssl_verify = args.ssl_verify

    # Return arguments
    return {"dest_dir": dest_dir, "ssl_verify": ssl_verify}


def load_sampleprojects(dest_dir: Path, ssl_verify: Optional[bool] = None):
    """Load the orthoseg sample projects.

    Args:
        dest_dir (Path): directory to save them to.
        ssl_verify (Optional[bool], optional): True or None to use the default
            certificate bundle as installed on your system. False disables certificate
            validation (NOT recommended!). If a path to a certificate bundle file (.pem)
            is passed, this will be used. In corporate networks using a proxy server
            this is often needed to evade CERTIFICATE_VERIFY_FAILED errors.
            Defaults to None.
    """
    dest_dir_full = dest_dir / "sample_projects"
    if dest_dir_full.exists():
        raise ValueError(f"Destination directory already exists: {dest_dir_full}")

    # Download
    print(f"Start download of sample projects to {dest_dir_full!s}")
    git_downloader.download(
        repo_url="https://github.com/orthoseg/orthoseg/tree/main/sample_projects",
        output_dir=dest_dir,
        ssl_verify=ssl_verify,
    )
    print("Download finished")
    print("Start download of footballfields pretrained neural net")
    verify = True if ssl_verify is None else ssl_verify
    footballfields_model_dir = dest_dir_full / "footballfields/models"
    footballfields_model_dir.mkdir(parents=True, exist_ok=True)
    model_hdf5_path = footballfields_model_dir / "footballfields_01_0.92512_242.hdf5"
    if model_hdf5_path.exists() is False:
        gdown.download(
            id="1XmAenCW6K_RVwqC6xbkapJ5ws-f7-QgH",
            output=str(model_hdf5_path),
            verify=verify,
        )
    model_hyperparams_path = (
        footballfields_model_dir / "footballfields_01_hyperparams.json"
    )
    if model_hyperparams_path.exists() is False:
        gdown.download(
            id="1umxcd4RkB81sem9PdIpLoWeiIW8ga1u7",
            output=str(model_hyperparams_path),
            verify=verify,
        )
    model_modeljson_path = footballfields_model_dir / "footballfields_01_model.json"
    if model_modeljson_path.exists() is False:
        gdown.download(
            id="16qe8thBTrO3dFfLMU1T22gWcfHVXt8zQ",
            output=str(model_modeljson_path),
            verify=verify,
        )
    print("Download finished")


def main():
    """Run load sampleprojects."""
    try:
        parsed_args = _parse_load_sampleprojects_args(sys.argv[1:])
        load_sampleprojects(**parsed_args)
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise


if __name__ == "__main__":
    main()



"""
Tests for functionalities in vector_util.
"""

import os

import geopandas as gpd
import pandas as pd
import pytest
from geopandas.testing import assert_geodataframe_equal
from pandas.testing import assert_frame_equal
from shapely import geometry as sh_geom

# Make hdf5 version warning non-blocking
os.environ["HDF5_DISABLE_VERSION_CHECK"] = "1"

from orthoseg.util import vector_util


@pytest.mark.parametrize("onborder_column_name", ["onborder_custom", "default_value"])
def test_is_onborder(onborder_column_name: str):
    # Prepare test data
    border_bounds = (0.0, 0, 20, 20)
    expected_onborder_column_name = "expected_onborder"
    assert expected_onborder_column_name != onborder_column_name
    columns = ["descr", "geometry", expected_onborder_column_name]
    testdata = [
        ["Geom None", None, 0],
        ["Geom None", None, 0],
        ["to be removed", None, 0],
        ["Polygon EMPTY", sh_geom.Polygon(), 0],
        [
            "polygon on border",
            sh_geom.Polygon(shell=[(10, 10), (10, 20), (20, 20), (20, 10), (10, 10)]),
            1,
        ],
        [
            "polygon not on border",
            sh_geom.Polygon(shell=[(10, 10), (10, 11), (11, 11), (11, 10), (10, 10)]),
            0,
        ],
    ]
    df = pd.DataFrame(testdata, columns=columns)
    testdata_gdf = gpd.GeoDataFrame(df, geometry="geometry")
    # Remove a row to test if the index is properly maintained
    testdata_gdf = testdata_gdf.drop([2], axis=0)
    assert isinstance(testdata_gdf, gpd.GeoDataFrame)

    if onborder_column_name == "default_value":
        result_gdf = vector_util.is_onborder(testdata_gdf, border_bounds=border_bounds)
        onborder_column_name = "onborder"
    else:
        result_gdf = vector_util.is_onborder(
            testdata_gdf,
            border_bounds=border_bounds,
            onborder_column_name=onborder_column_name,
        )

    # Compare result with expected result
    assert onborder_column_name in result_gdf.columns
    assert onborder_column_name not in testdata_gdf.columns
    result_gdf[expected_onborder_column_name] = result_gdf[onborder_column_name]
    result_gdf = result_gdf.drop(columns=[onborder_column_name])
    assert_geodataframe_equal(testdata_gdf, result_gdf)

    # Test empty GeoDataFrame
    # --------------------
    result_gdf = vector_util.is_onborder(
        gpd.GeoDataFrame(),
        border_bounds=border_bounds,
        onborder_column_name=onborder_column_name,
    )
    assert result_gdf is not None
    assert len(result_gdf) == 0


def test_reclassify_neighbours():
    # Prepare test data
    reclassify_column = "class_name"
    query = "area <= 20 and onborder == 0"
    border_bounds = (0.0, 0, 20, 20)
    columns = ["test_id", "descr", "expected_result", "geometry", reclassify_column]
    testdata = [
        ["None1", "Geom None", "disappears", None, "class_1"],
        ["None2", "Geom None", "disappears", None, "class_1"],
        ["None3", "to be removed", "disappears", None, "class_1"],
        ["Empty1", "Polygon EMPTY", "disappears", sh_geom.Polygon(), "class_1"],
        [
            "poly1",
            "large polygon (25m²)",
            "smaller polygons are added to it",
            sh_geom.Polygon(shell=[(5, 5), (10, 5), (10, 10), (5, 10), (5, 5)]),
            "class_1_large",
        ],
        [
            "poly2",
            "medium (5m²), neighbour of polygon1, onborder",
            "is onborder, so isn't merged with other things",
            sh_geom.Polygon(shell=[(0, 5), (0, 6), (5, 6), (5, 5), (0, 5)]),
            "class_2_medium",
        ],
        [
            "poly3",
            "small (1m²), neighbour of poly1 and poly4, not onborder",
            "is merged with poly1",
            sh_geom.Polygon(shell=[(10, 5), (11, 5), (11, 6), (10, 6), (10, 5)]),
            "class_3_small",
        ],
        [
            "poly4",
            "small (1m²), neighbour of poly3, not onborder",
            "is merged with poly1 and poly3",
            sh_geom.Polygon(shell=[(11, 5), (12, 5), (12, 6), (11, 6), (11, 5)]),
            "class_4_small",
        ],
        [
            "poly5",
            "medium polygon (5m²), larger than neighbour poly6, not onborder",
            "is merged with poly6 but retains class",
            sh_geom.Polygon(shell=[(5, 15), (10, 15), (10, 16), (5, 16), (5, 15)]),
            "class_5_medium",
        ],
        [
            "poly6",
            "small (1m²), smaller than neighbour poly5, not onborder",
            "is merged with poly6 and gets its class",
            sh_geom.Polygon(shell=[(4, 15), (5, 15), (5, 16), (4, 16), (4, 15)]),
            "class_6_small",
        ],
    ]
    df = pd.DataFrame(testdata, columns=columns)
    testdata_gdf = gpd.GeoDataFrame(df, geometry="geometry")
    # Remove a row to test if the index is properly maintained
    testdata_gdf = testdata_gdf.drop([2], axis=0)

    expected_result_df = pd.DataFrame(
        [
            ["poly1, poly3, poly4", "class_1_large"],
            ["poly2", "class_2_medium"],
            ["poly5, poly6", "class_5_medium"],
        ],
        columns=["test_id", reclassify_column],
    )
    # Test with extra column, values are concatenated
    # -----------------------------------------------
    # reclassify only supported on gdf with only the reclassify column
    testdata_prepared_gdf = testdata_gdf[["test_id", "geometry", reclassify_column]]
    assert isinstance(testdata_prepared_gdf, gpd.GeoDataFrame)
    result_gdf = vector_util.reclassify_neighbours(
        testdata_prepared_gdf, reclassify_column, query, border_bounds
    )
    assert result_gdf is not None
    result_df = pd.DataFrame(result_gdf[["test_id", reclassify_column]])
    assert_frame_equal(result_df, expected_result_df)

    # Test with no extra column
    # -------------------------
    expected_result_df = expected_result_df[[reclassify_column]]

    # reclassify only supported on gdf with only the reclassify column
    testdata_prepared_gdf = testdata_gdf[["geometry", reclassify_column]]
    assert isinstance(testdata_prepared_gdf, gpd.GeoDataFrame)
    result_gdf = vector_util.reclassify_neighbours(
        testdata_prepared_gdf, reclassify_column, query, border_bounds
    )
    assert result_gdf is not None
    result_df = pd.DataFrame(result_gdf[[reclassify_column]])
    assert_frame_equal(result_df, expected_result_df)



from contextlib import AbstractContextManager, nullcontext

import pytest

from orthoseg.util.data import aidetection_info


@pytest.mark.parametrize(
    "file, exp_error",
    [
        ("footballfields_01_201_BEFL-2019_dissolve.gpkg", False),
        ("footballfields_01_201.gpkg", True),
    ],
)
def test_aidetection_info(tmp_path, file: str, exp_error: bool):
    # Create test project
    testproject_dir = tmp_path / "orthoseg_test_aidetection_info"
    testproject_dir.mkdir(parents=True, exist_ok=True)
    path = testproject_dir / file
    path.touch()

    handler: AbstractContextManager
    if exp_error:
        matchstr = r"Error in get_aidetection_info on .*"
        handler = pytest.raises(ValueError, match=matchstr)
    else:
        handler = nullcontext()

    with handler:
        # Get aidetection info
        ai_detection_info = aidetection_info(path=path)

        assert ai_detection_info.path == path
        assert ai_detection_info.subject == "footballfields"
        assert ai_detection_info.traindata_version == 1
        assert ai_detection_info.image_layer == "BEFL-2019"
        assert ai_detection_info.image_layer_year == 2019
        assert ai_detection_info.postprocessing == "dissolve"



"""Module with functions for post-processing prediction masks towards polygons."""

import argparse
import logging
import shlex
import shutil
import sys
import traceback
from pathlib import Path

import orthoseg.model.model_helper as mh
from orthoseg.helpers import config_helper as conf, email_helper
from orthoseg.lib import postprocess_predictions as postp
from orthoseg.util import log_util

# Get a logger...
logger = logging.getLogger(__name__)


def _postprocess_argstr(argstr):
    args = shlex.split(argstr)
    _postprocess_args(args)


def _postprocess_args(args):
    # Interprete arguments
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group("Required arguments")
    required.add_argument(
        "-c", "--config", type=str, required=True, help="The config file to use"
    )

    # Optional arguments
    optional = parser.add_argument_group("Optional arguments")
    # Add back help
    optional.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )
    optional.add_argument(
        "config_overrules",
        nargs="*",
        help=(
            "Supply any number of config overrules like this: "
            "<section>.<parameter>=<value>"
        ),
    )

    # Interprete arguments
    args = parser.parse_args(args)

    # Run!
    postprocess(config_path=Path(args.config), config_overrules=args.config_overrules)


def postprocess(config_path: Path, config_overrules: list[str] = []):
    """Postprocess the output of a prediction for the config specified.

    Args:
        config_path (Path): Path to the config file.
        config_overrules (list[str], optional): list of config options that will
            overrule other ways to supply configuration. They should be specified in the
            form of "<section>.<parameter>=<value>". Defaults to [].
    """
    # Init
    # Load the config and save in a bunch of global variables zo it
    # is accessible everywhere
    conf.read_orthoseg_config(config_path, overrules=config_overrules)

    # Init logging
    log_util.clean_log_dir(
        log_dir=conf.dirs.getpath("log_dir"),
        nb_logfiles_tokeep=conf.logging_conf.getint("nb_logfiles_tokeep"),
    )
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath("log_dir"), __name__)

    # Log start + send email
    message = f"Start postprocess for config {config_path.stem}"
    logger.info(message)
    logger.debug(f"Config used: \n{conf.pformat_config()}")
    email_helper.sendmail(message)

    try:
        # Create base filename of model to use
        # TODO: is force data version the most logical, or rather implement
        #       force weights file or ?
        traindata_id = None
        force_model_traindata_id = conf.train.getint("force_model_traindata_id")
        if force_model_traindata_id is not None and force_model_traindata_id > -1:
            traindata_id = force_model_traindata_id

        # Get the best model that already exists for this train dataset
        trainparams_id = conf.train.getint("trainparams_id")
        best_model = mh.get_best_model(
            model_dir=conf.dirs.getpath("model_dir"),
            segment_subject=conf.general["segment_subject"],
            traindata_id=traindata_id,
            trainparams_id=trainparams_id,
        )
        if best_model is None:
            raise RuntimeError(
                f"No best model found in {conf.dirs.getpath('model_dir')}"
            )

        # Input file  the "most recent" prediction result dir for this subject
        output_vector_dir = conf.dirs.getpath("output_vector_dir")
        output_vector_name = (
            f"{best_model['basefilename']}_{best_model['epoch']}_"
            f"{conf.predict['image_layer']}"
        )
        output_vector_path = output_vector_dir / f"{output_vector_name}.gpkg"

        # Prepare some parameters for the postprocessing
        nb_parallel = conf.general.getint("nb_parallel", -1)

        keep_original_file = conf.postprocess.getboolean("keep_original_file", True)
        keep_intermediary_files = conf.postprocess.getboolean(
            "keep_intermediary_files", True
        )
        dissolve = conf.postprocess.getboolean("dissolve", True)
        dissolve_tiles_path = conf.postprocess.getpath("dissolve_tiles_path")
        reclassify_query = conf.postprocess.get("reclassify_to_neighbour_query")
        if reclassify_query is not None:
            reclassify_query = reclassify_query.replace("\n", " ")

        simplify_algorithm = conf.postprocess.get("simplify_algorithm")
        simplify_tolerance = conf.postprocess.geteval("simplify_tolerance")
        simplify_lookahead = conf.postprocess.get("simplify_lookahead")
        if simplify_lookahead is not None:
            simplify_lookahead = int(simplify_lookahead)

        # Go!
        postp.postprocess_predictions(
            input_path=output_vector_path,
            output_path=output_vector_path,
            keep_original_file=keep_original_file,
            keep_intermediary_files=keep_intermediary_files,
            dissolve=dissolve,
            dissolve_tiles_path=dissolve_tiles_path,
            reclassify_to_neighbour_query=reclassify_query,
            simplify_algorithm=simplify_algorithm,
            simplify_tolerance=simplify_tolerance,
            simplify_lookahead=simplify_lookahead,
            nb_parallel=nb_parallel,
            force=False,
        )

        # Log and send mail
        message = f"Completed postprocess for config {config_path.stem}"
        logger.info(message)
        email_helper.sendmail(message)
    except Exception as ex:
        message = f"ERROR while running postprocess for task {config_path.stem}"
        logger.exception(message)
        email_helper.sendmail(
            subject=message, body=f"Exception: {ex}\n\n {traceback.format_exc()}"
        )
        raise RuntimeError(message) from ex
    finally:
        if conf.tmp_dir is not None:
            shutil.rmtree(conf.tmp_dir, ignore_errors=True)


def main():
    """Run postprocess."""
    try:
        _postprocess_args(sys.argv[1:])
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise


# If the script is ran directly...
if __name__ == "__main__":
    main()



"""
Tests for functionalities in orthoseg.train.
"""

import os
import platform
import shutil
import sys
from pathlib import Path

import pytest

from orthoseg import load_sampleprojects


def get_testdata_dir() -> Path:
    return Path(__file__).resolve().parent / "data"


@pytest.mark.skipif(
    "GITHUB_ACTIONS" in os.environ
    and not (
        platform.system() == "Linux"
        and sys.version_info.major == 3
        and sys.version_info.minor == 10
    ),
    reason="on github CI, run this only in one env to avoid rate limit exceeded",
)
def test_load_sampleprojects(tmp_path):
    sampleprojects_dir = tmp_path / "sample_projects"
    shutil.rmtree(sampleprojects_dir, ignore_errors=True)
    load_sampleprojects.load_sampleprojects(dest_dir=sampleprojects_dir.parent)

    # Check if the files were correctly loaded
    assert sampleprojects_dir.exists()
    assert (sampleprojects_dir / "imagelayers.ini").exists()
    assert (sampleprojects_dir / "project_defaults_overrule.ini").exists()
    assert (sampleprojects_dir / "run_footballfields.py").exists()

    footballfields_dir = sampleprojects_dir / "footballfields"
    assert footballfields_dir.exists()
    files = list((footballfields_dir).glob("**/*.ini"))
    assert len(files) > 0
    files = list((footballfields_dir / "labels").glob("**/*.gpkg"))
    assert len(files) == 2
    model_path = footballfields_dir / "models" / "footballfields_01_0.92512_242.hdf5"
    assert model_path.exists()
    # The model should be larger than 50 MB, otherwise not normal
    assert model_path.stat().st_size > 50 * 1024 * 1024

    projecttemplate_dir = sampleprojects_dir / "project_template"
    assert projecttemplate_dir.exists()
    files = list((projecttemplate_dir).glob("**/*.ini"))
    assert len(files) > 0
    files = list((projecttemplate_dir / "labels").glob("**/*.gpkg"))
    assert len(files) == 2



"""Module with helper functions to create models.

Offers a common interface, regardless of the underlying model implementation
and contains extra metrics, callbacks,...

Many models are supported by using this segmentation model zoo:
https://github.com/qubvel/segmentation_models
"""

# Disable isort, another import order gives a segmentation fault on Windows.
import json  # noqa: I001
import logging
import os
from pathlib import Path
from typing import Any, Callable, Optional, Union

import numpy as np
import tensorflow as tf
import keras.models

# Get a logger...
logger = logging.getLogger(__name__)
os.environ["SM_FRAMEWORK"] = "tf.keras"

"""
preprocessing_fn = get_preprocessing('resnet34')
x = preprocessing_fn(x)
"""


def get_model(
    architecture: str,
    input_width: Optional[int] = None,
    input_height: Optional[int] = None,
    nb_channels: int = 3,
    nb_classes: int = 1,
    activation: str = "softmax",
    init_weights_with: str = "imagenet",
    freeze: bool = False,
) -> keras.models.Model:
    """Get a model.

    Args:
        architecture (str): Architecture of the network to create
        input_width (int, optional): Width of the input images. Defaults to None.
        input_height (int, optional): Height of the input images. Defaults to None.
        nb_channels (int, optional): Nb of channels/bands of the input images.
            Defaults to 3.
        nb_classes (int, optional): Nb of classes to be segmented to. Defaults to 1.
        activation (Activation, optional): Activation function of last layer.
            Defaults to 'softmax'.
        init_weights_with (str, optional): Weights to init the network with.
            Defaults to 'imagenet'.
        freeze (bool, optional): Freeze the final layer weights during
            training. It is usefull to use this option for the first few
            epochs get a more robust network. Defaults to False.

    Returns:
        Model: the model.
    """
    # Check architecture
    segment_architecture_parts = architecture.split("+")
    if len(segment_architecture_parts) < 2:
        raise Exception(f"Unsupported architecture: {architecture}")
    encoder = segment_architecture_parts[0]
    decoder = segment_architecture_parts[1]

    if decoder.lower() == "unet":
        # Architecture implemented using the segmentation_models library
        from segmentation_models import Unet

        model = Unet(
            backbone_name=encoder.lower(),
            input_shape=(input_width, input_height, nb_channels),
            classes=nb_classes,
            activation=activation,
            encoder_weights=init_weights_with,
            encoder_freeze=freeze,
        )
        return model

    elif decoder.lower() == "pspnet":
        # Architecture implemented using the segmentation_models library
        from segmentation_models import PSPNet

        model = PSPNet(
            backbone_name=encoder.lower(),
            input_shape=(input_width, input_height, nb_channels),
            classes=nb_classes,
            activation=activation,
            encoder_weights=init_weights_with,
            encoder_freeze=freeze,
        )
        return model

    elif decoder.lower() == "linknet":
        # Architecture implemented using the segmentation_models library
        from segmentation_models import Linknet

        # First check if input size is compatible with linknet
        if input_width is not None and input_height is not None:
            check_image_size(architecture, input_width, input_height)

        model = Linknet(
            backbone_name=encoder.lower(),
            input_shape=(input_width, input_height, nb_channels),
            classes=nb_classes,
            activation=activation,
            encoder_weights=init_weights_with,
            encoder_freeze=freeze,
        )
        return model

    else:
        raise ValueError(f"Unknown decoder architecture: {decoder}")


def compile_model(
    model: keras.models.Model,
    optimizer: str,
    optimizer_params: dict,
    loss: str,
    metrics: Optional[list[str]] = None,
    sample_weight_mode: Optional[str] = None,
    class_weights: Optional[list] = None,
) -> keras.models.Model:
    """Compile the model for training.

    Args:
        model (keras.models.Model): the keras model to compile.
        optimizer (str): the optimizer to use.
        optimizer_params (dict): parameters to use for optimizer.
        loss (str): the loss function to use. One of:
            * categorical_crossentropy
            * weighted_categorical_crossentropy: class_weights should be specified!

        metrics (list[Metric], optional): metrics to use. Defaults to None. One of:
            *
        sample_weight_mode (str, optional): sample weight mode to use. Defaults to None.
        class_weights (list, optional): class weigths to use. Defaults to None.
    """
    import segmentation_models

    # Get number classes of model
    nb_classes = model.output[-1].shape[-1]

    # If no metrics specified, use default ones
    metric_funcs: list[Any] = []
    if metrics is None:
        if loss in ["categorical_crossentropy", "weighted_categorical_crossentropy"]:
            metric_funcs.append("categorical_accuracy")
        elif loss == "sparse_categorical_crossentropy":
            # metrics.append('sparse_categorical_accuracy')
            logger.warning("loss == sparse_categorical_crossentropy not implementedd")
        elif loss == "binary_crossentropy":
            metric_funcs.append("binary_accuracy")

        onehot_mean_iou = tf.keras.metrics.OneHotMeanIoU(
            num_classes=nb_classes, name="one_hot_mean_iou"
        )
        metric_funcs.append(onehot_mean_iou)
    else:
        raise ValueError("Specifying metrics not yet implemented")

    # Check loss function
    loss_func: Union[Callable, str]
    if loss == "bcedice":
        loss_func = dice_coef_loss_bce
    elif loss == "dice_loss":
        loss_func = segmentation_models.losses.DiceLoss()
    elif loss == "jaccard_loss":
        loss_func = segmentation_models.losses.JaccardLoss()
    elif loss == "weighted_categorical_crossentropy":
        if class_weights is None:
            raise ValueError(f"With loss == {loss}, class_weights cannot be None!")
        loss_func = weighted_categorical_crossentropy(class_weights)
    else:
        loss_func = loss

    # Create optimizer
    if optimizer == "adam":
        optimizer_func = tf.keras.optimizers.Adam(**optimizer_params)
    else:
        raise ValueError(
            f"Error creating optimizer: {optimizer}, with params {optimizer_params}"
        )

    logger.info(
        f"Compile model, optimizer: {optimizer}, loss: {loss}, "
        f"class_weights: {class_weights}"
    )
    model.compile(
        optimizer=optimizer_func,
        loss=loss_func,
        metrics=metric_funcs,
        sample_weight_mode=sample_weight_mode,
    )

    return model


def load_model(model_to_use_filepath: Path, compile: bool = True) -> keras.models.Model:
    """Load an existing model from a file.

    If loading the architecture + model from the file doesn't work, tries
    different things to load a model anyway:
    1) looks for a ..._model.json file to load the architecture from,
       the weights are then loaded from the original file
    2) looks for a ..._hyperparams.json file to create the architecture from,
       the weights are then loaded from the original file

    Args:
        model_to_use_filepath (Path): The file to load the model from.
        compile (bool, optional): True to get a compiled version back that is
            ready to train. Defaults to True.

    Raises:
        Exception: [description]

    Returns:
        tf.keras.models.Model: The loaded model.
    """
    import segmentation_models

    errors = []
    model = None
    model_basestem = f"{'_'.join(model_to_use_filepath.stem.split('_')[0:2])}"

    # If it is a file with the complete model, try loading it entirely...
    if not model_to_use_filepath.stem.endswith("_weights"):
        # Load the hyperparams file
        hyperparams_json_filepath = (
            model_to_use_filepath.parent / f"{model_basestem}_hyperparams.json"
        )
        nb_classes = 1
        if hyperparams_json_filepath.exists():
            with hyperparams_json_filepath.open("r") as src:
                hyperparams = json.load(src)
                nb_classes = len(hyperparams["architecture"]["classes"])

        iou_score = segmentation_models.metrics.IOUScore()
        f1_score = segmentation_models.metrics.FScore()
        onehot_mean_iou = tf.keras.metrics.OneHotMeanIoU(
            num_classes=nb_classes, name="one_hot_mean_iou"
        )

        try:
            model = tf.keras.models.load_model(
                str(model_to_use_filepath),
                custom_objects={
                    "jaccard_coef": jaccard_coef,
                    "jaccard_coef_flat": jaccard_coef_flat,
                    "jaccard_coef_round": jaccard_coef_round,
                    "dice_coef": dice_coef,
                    "iou_score": iou_score,
                    "f1_score": f1_score,
                    "one_hot_mean_iou": onehot_mean_iou,
                    "weighted_categorical_crossentropy": weighted_categorical_crossentropy,  # noqa: E501
                },
                compile=compile,
            )
        except Exception as ex:
            errors.append(
                f"Error loading model+weights from {model_to_use_filepath}: {ex}"
            )

    # If no model returned yet, try loading loading architecture and weights seperately
    if model is None:
        # Load the architecture from a model.json file
        # Check if there is a specific model.json file for these weights
        model_json_filepath = model_to_use_filepath.parent / (
            model_to_use_filepath.stem.replace("_weights", "") + ".json"
        )
        if not model_json_filepath.exists():
            # If not, check if there is a model.json file for the training session
            model_json_filepath = (
                model_to_use_filepath.parent / f"{model_basestem}_model.json"
            )
        if model_json_filepath.exists():
            with model_json_filepath.open("r") as src:
                model_json = src.read()
            try:
                model = tf.keras.models.model_from_json(model_json)
            except Exception as ex:
                errors.append(
                    f"Error loading model architecture from {model_json_filepath}: {ex}"
                )
        else:
            errors.append(
                f"No model.json file found to load model from: {model_json_filepath}"
            )

        # If loading model.json not successfull, create model based on hyperparams.json
        if model is None:
            # Load the hyperparams file if available
            hyperparams_json_filepath = (
                model_to_use_filepath.parent / f"{model_basestem}_hyperparams.json"
            )
            if hyperparams_json_filepath.exists():
                with hyperparams_json_filepath.open("r") as src:
                    hyperparams = json.load(src)
                # Create the model we want to use
                try:
                    model = get_model(
                        architecture=hyperparams["architecture"]["architecture"],
                        nb_channels=hyperparams["architecture"]["nb_channels"],
                        nb_classes=hyperparams["architecture"]["nb_classes"],
                        activation=hyperparams["architecture"]["activation_function"],
                    )
                except Exception as ex:
                    errors.append(
                        "Error in get_model() with params from: "
                        f"{hyperparams_json_filepath}: {ex}"
                    )
            else:
                errors.append(
                    f"No file found to load model from: {hyperparams_json_filepath}"
                )

        # Now load the weights
        if model is not None:
            try:
                model.load_weights(str(model_to_use_filepath))
            except Exception as ex:
                errors.append(
                    f"Error trying model.load_weights on: {model_to_use_filepath}: {ex}"
                )

    # Check if a model got loaded...
    if model is not None:
        # Eager seems to be 50% slower, tested on tensorflow 2.5
        model.run_eagerly = False
    else:
        # If we still have not model... time to give up.
        errors_str = ""
        if len(errors) > 0:
            errors_str = (
                " The following errors occured while trying: \n    -> "
                + "\n    -> ".join(errors)
            )
        raise Exception(f"Error loading model for {model_to_use_filepath}.{errors_str}")

    return model


def set_trainable(model, recompile: bool = True):
    """Set the model trainable.

    Args:
        model (_type_): model to set trainable.
        recompile (bool, optional): True to recompile the model so it is ready to train.
            Defaults to True.
    """
    import segmentation_models

    # doesn't seem to work, so save and load model
    segmentation_models.utils.set_trainable(model=model, recompile=recompile)


def check_image_size(architecture: str, input_width: int, input_height: int):
    """Check if the image size is compatible with the architecture.

    A ValueError is raised if the architecture is not compatible with the size.

    Args:
        architecture (str): architecture to check compatibility for.
        input_width (int): image width.
        input_height (int): image height.
    """
    # Check architecture
    segment_architecture_parts = architecture.split("+")
    if len(segment_architecture_parts) < 2:
        raise ValueError(f"Unsupported architecture: {architecture}")
    # encoder = segment_architecture_parts[0]
    decoder = segment_architecture_parts[1]

    if decoder.lower() == "linknet":
        if (input_width and (input_width % 32) != 0) or (
            input_height and (input_height % 32) != 0
        ):
            message = (
                f"for decoder linknet, input image width ({input_width}) and "
                f"input image height ({input_height}) should be divisible by 32!"
            )
            logger.error(message)
            raise ValueError(message)


# ------------------------------------------
# Loss functions
# ------------------------------------------


def weighted_categorical_crossentropy(weights):
    """Loss function using weighted categorical crossentropy.

    Args:
        weights (ktensor|nparray|list): crossentropy weights
    Returns:
        weighted categorical crossentropy function
    """
    if isinstance(weights, (list, np.ndarray)):
        weights = tf.keras.backend.variable(weights)

    def loss(target, output, from_logits=False):
        if not from_logits:
            output /= tf.reduce_sum(output, len(output.get_shape()) - 1, True)
            _epsilon = tf.convert_to_tensor(
                tf.keras.backend.epsilon(), dtype=output.dtype.base_dtype
            )
            output = tf.clip_by_value(output, _epsilon, 1.0 - _epsilon)
            weighted_losses = target * tf.math.log(output) * weights
            retval = -tf.reduce_sum(weighted_losses, len(output.get_shape()) - 1)
            return retval
        else:
            raise ValueError("WeightedCategoricalCrossentropy: not valid with logits")

    return loss


def dice_coef_loss(y_true, y_pred):
    """Loss function based of dice coefficient.

    Args:
        y_true (_type_): _description_
        y_pred (_type_): _description_

    Returns:
        _type_: _description_
    """
    return 1 - dice_coef(y_true, y_pred)


def bootstrapped_crossentropy(y_true, y_pred, bootstrap_type="hard", alpha=0.95):
    """Loss function based on cross entropy with a bootstrap.

    Args:
        y_true (_type_): _description_
        y_pred (_type_): _description_
        bootstrap_type (str, optional): _description_. Defaults to "hard".
        alpha (float, optional): _description_. Defaults to 0.95.

    Returns:
        _type_: _description_
    """
    target_tensor = y_true
    prediction_tensor = y_pred
    _epsilon = tf.convert_to_tensor(
        tf.keras.backend.epsilon(), prediction_tensor.dtype.base_dtype
    )
    prediction_tensor = tf.clip_by_value(prediction_tensor, _epsilon, 1 - _epsilon)
    prediction_tensor = tf.keras.backend.log(
        prediction_tensor / (1 - prediction_tensor)
    )

    if bootstrap_type == "soft":
        bootstrap_target_tensor = alpha * target_tensor + (1.0 - alpha) * tf.sigmoid(
            prediction_tensor
        )
    else:
        bootstrap_target_tensor = alpha * target_tensor + (1.0 - alpha) * tf.cast(
            tf.sigmoid(prediction_tensor) > 0.5, tf.float32
        )
    return tf.keras.backend.mean(
        tf.nn.sigmoid_cross_entropy_with_logits(
            labels=bootstrap_target_tensor, logits=prediction_tensor
        )
    )


def dice_coef_loss_bce(y_true, y_pred):
    """Loss function based on dice coefficient with bootstrapping.

    Args:
        y_true (_type_): _description_
        y_pred (_type_): _description_

    Returns:
        _type_: _description_
    """
    dice = 0.5
    bce = 0.5
    bootstrapping = "hard"
    alpha = 1.0
    return (
        bootstrapped_crossentropy(y_true, y_pred, bootstrapping, alpha) * bce
        + dice_coef_loss(y_true, y_pred) * dice
    )


# ------------------------------------------
# Metrics functions
# ------------------------------------------

SMOOTH_LOSS = 1e-12


def jaccard_coef(y_true, y_pred):
    """Metric jaccard coefficient aka intersection over union.

    Args:
        y_true (_type_): _description_
        y_pred (_type_): _description_

    Returns:
        _type_: _description_
    """
    intersection = tf.keras.backend.sum(y_true * y_pred, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred, axis=[0, -1, -2])

    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)

    return tf.keras.backend.mean(jac)


def jaccard_coef_round(y_true, y_pred):
    """Metric jaccard coefficient aka intersection over union with rounding.

    Args:
        y_true (_type_): _description_
        y_pred (_type_): _description_

    Returns:
        _type_: _description_
    """
    y_pred_pos = tf.keras.backend.round(tf.keras.backend.clip(y_pred, 0, 1))

    intersection = tf.keras.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)
    return tf.keras.backend.mean(jac)


def jaccard_coef_flat(y_true, y_pred):
    """Metric jaccard coefficient aka intersection over union with flattening.

    Args:
        y_true (_type_): _description_
        y_pred (_type_): _description_

    Returns:
        _type_: _description_
    """
    y_true_f = tf.keras.backend.flatten(y_true)
    y_pred_f = tf.keras.backend.flatten(y_pred)
    intersection = tf.keras.backend.sum(y_true_f * y_pred_f)
    return (intersection + SMOOTH_LOSS) / (
        tf.keras.backend.sum(y_true_f)
        + tf.keras.backend.sum(y_pred_f)
        - intersection
        + SMOOTH_LOSS
    )


def dice_coef(y_true, y_pred, smooth=1.0):
    """Metric dice coefficient.

    Args:
        y_true (_type_): _description_
        y_pred (_type_): _description_
        smooth (float, optional): _description_. Defaults to 1.0.

    Returns:
        _type_: _description_
    """
    y_true_f = tf.keras.backend.flatten(y_true)
    y_pred_f = tf.keras.backend.flatten(y_pred)
    intersection = tf.keras.backend.sum(y_true_f * y_pred_f)
    return (2.0 * intersection + smooth) / (
        tf.keras.backend.sum(y_true_f) + tf.keras.backend.sum(y_pred_f) + smooth
    )


def pct_wrong(y_true, y_pred):
    """Metric percentage wrong.

    Args:
        y_true (_type_): _description_
        y_pred (_type_): _description_

    Returns:
        _type_: _description_
    """
    y_pred_pos = tf.keras.backend.round(tf.keras.backend.clip(y_pred, 0, 1))

    intersection = tf.keras.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)
    return tf.keras.backend.mean(jac)



"""High-level API to run a segmentation."""

import argparse
import logging
import pprint
import shutil
import sys
import traceback
from pathlib import Path
from typing import Any

# import os
# os.environ["CUDA_VISIBLE_DEVICES"] = "-1" # Disable using GPU
import tensorflow as tf

import orthoseg.model.model_factory as mf
import orthoseg.model.model_helper as mh
from orthoseg.helpers import config_helper as conf, email_helper
from orthoseg.lib import cleanup, predicter
from orthoseg.util import log_util

# Get a logger...
logger = logging.getLogger(__name__)


def _predict_args(args) -> argparse.Namespace:
    # Interprete arguments
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group("Required arguments")
    required.add_argument(
        "-c", "--config", type=str, required=True, help="The config file to use"
    )

    # Optional arguments
    optional = parser.add_argument_group("Optional arguments")
    # Add back help
    optional.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )
    optional.add_argument(
        "config_overrules",
        nargs="*",
        help=(
            "Supply any number of config overrules like this: "
            "<section>.<parameter>=<value>"
        ),
    )

    return parser.parse_args(args)


def predict(config_path: Path, config_overrules: list[str] = []):
    """Run a prediction for the config specified.

    Args:
        config_path (Path): Path to the config file to use.
        config_overrules (list[str], optional): list of config options that will
            overrule other ways to supply configuration. They should be specified in the
            form of "<section>.<parameter>=<value>". Defaults to [].
    """
    # Init
    # Load the config and save in a bunch of global variables zo it
    # is accessible everywhere
    conf.read_orthoseg_config(config_path, overrules=config_overrules)

    # Init logging
    log_util.clean_log_dir(
        log_dir=conf.dirs.getpath("log_dir"),
        nb_logfiles_tokeep=conf.logging_conf.getint("nb_logfiles_tokeep"),
    )
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath("log_dir"), __name__)
    logger.info(f"Start predict for config {config_path.stem}")
    logger.debug(f"Config used: \n{conf.pformat_config()}")

    try:
        # Read some config, and check if values are ok
        image_layer = conf.predict["image_layer"]
        image_layer_info = conf.image_layers[image_layer]
        if image_layer_info is None:
            raise Exception(
                f"image_layer to predict is not specified in config: {image_layer_info}"
            )
        input_image_dir = conf.dirs.getpath("predict_image_input_dir")
        if not input_image_dir.exists():
            raise Exception(f"input image dir doesn't exist: {input_image_dir}")

        # Create base filename of model to use
        # TODO: is force data version the most logical, or rather implement
        #       force weights file or ?
        traindata_id = None
        force_model_traindata_id = conf.train.getint("force_model_traindata_id")
        if force_model_traindata_id is not None and force_model_traindata_id > -1:
            traindata_id = force_model_traindata_id

        # Get the best model that already exists for this train dataset
        trainparams_id = conf.train.getint("trainparams_id")
        best_model = mh.get_best_model(
            model_dir=conf.dirs.getpath("model_dir"),
            segment_subject=conf.general["segment_subject"],
            traindata_id=traindata_id,
            trainparams_id=trainparams_id,
        )

        # Check if a model was found
        if best_model is None:
            message = (
                f"No model found in model_dir: {conf.dirs.getpath('model_dir')} for "
                f"traindata_id: {traindata_id}"
            )
            logger.critical(message)
            raise Exception(message)
        else:
            model_weights_filepath = best_model["filepath"]
            logger.info(f"Best model found: {model_weights_filepath}")

        # Load the hyperparams of the model
        # TODO: move the hyperparams filename formatting to get_models...
        hyperparams_path = (
            best_model["filepath"].parent
            / f"{best_model['basefilename']}_hyperparams.json"
        )
        hyperparams = mh.HyperParams(path=hyperparams_path)

        # Validate the image prediction size for the model architecture
        overlap = conf.predict.getint("image_pixels_overlap", 0)
        input_width_pred = conf.predict.getint("image_pixel_width") + 2 * overlap
        input_height_pred = conf.predict.getint("image_pixel_height") + 2 * overlap
        mf.check_image_size(
            architecture=hyperparams.architecture.architecture,
            input_width=input_width_pred,
            input_height=input_height_pred,
        )

        # Prepare output subdir to be used for predictions
        predict_out_subdir = f"{best_model['basefilename']}"
        if trainparams_id > 0:
            predict_out_subdir += f"_{trainparams_id}"
        predict_out_subdir += f"_{best_model['epoch']}"

        # Load model to predict with
        # --------------------------
        model = None
        # Try optimizing model with tensorrt. Not supported on Windows
        # -> disabled for now till I have access again to a linux machine
        """
        if os.name != "nt":
            try:
                # Try import
                from tensorflow.python.compiler.tensorrt import trt_convert as trt

                # Import didn't fail, so optimize model
                logger.info(
                    "Tensorrt is available, so try to create and use optimized model"
                )
                savedmodel_optim_dir = (
                    best_model["filepath"].parent
                    / f"{best_model['filepath'].stem}_optim"
                )
                if not savedmodel_optim_dir.exists():
                    # If base model not yet in savedmodel format
                    savedmodel_dir = (
                        best_model["filepath"].parent / best_model["filepath"].stem
                    )
                    if not savedmodel_dir.exists():
                        logger.info(
                            f"SavedModel format not yet available, so load "
                            f"model + weights from {best_model['filepath']}"
                        )
                        model = mf.load_model(best_model["filepath"], compile=False)
                        logger.info(f"Now save again as savedmodel to {savedmodel_dir}")
                        tf.saved_model.save(model, str(savedmodel_dir))
                        model = None

                    # Now optimize model
                    logger.info(f"Optimize + save model to {savedmodel_optim_dir}")
                    converter = trt.TrtGraphConverterV2(
                        input_saved_model_dir=str(savedmodel_dir),
                        is_dynamic_op=True,
                        precision_mode="FP16",
                    )
                    converter.convert()
                    converter.save(savedmodel_optim_dir)

                logger.info(
                    f"Load optimized model + weights from {savedmodel_optim_dir}"
                )
                model = tf.keras.models.load_model(str(savedmodel_optim_dir))

            except ImportError:
                logger.info("Tensorrt is not available, so load unoptimized model")
            except Exception as ex:
                logger.info(
                    "An error occured trying to use tensorrt, "
                    f"so load unoptimized model. Error: {ex}"
                )
        """

        # If model isn't loaded yet... load!
        if model is None:
            model = mf.load_model(best_model["filepath"], compile=False)

        # Prepare the model for predicting
        nb_gpu = len(tf.config.experimental.list_physical_devices("GPU"))
        batch_size = conf.predict.getint("batch_size")
        if nb_gpu <= 1:
            model_for_predict = model
            logger.info(f"Predict using single GPU or CPU, with nb_gpu: {nb_gpu}")
        else:
            # If multiple GPU's available, create multi_gpu_model
            try:
                model_for_predict = model
                logger.warning(
                    "Predict using multiple GPUs NOT IMPLEMENTED AT THE MOMENT"
                )

                # logger.info(
                #     f"Predict using multiple GPUs: {nb_gpu}, batch size becomes: "
                #     f"{batch_size*nb_gpu}"
                # )
                # batch_size *= nb_gpu
            except ValueError:
                logger.info("Predict using single GPU or CPU")
                model_for_predict = model

        # Prepare params for the inline postprocessing of the prediction
        min_probability = conf.predict.getfloat("min_probability")
        postprocess: dict[str, Any] = {}
        simplify_algorithm = conf.predict.get("simplify_algorithm")
        if simplify_algorithm is not None and simplify_algorithm != (""):
            postprocess["simplify"] = {}
            simplify = postprocess["simplify"]

            simplify["simplify_algorithm"] = simplify_algorithm
            simplify["simplify_tolerance"] = conf.predict.geteval("simplify_tolerance")
            simplify["simplify_lookahead"] = conf.predict.getint("simplify_lookahead")
            simplify["simplify_topological"] = conf.predict.getboolean_ext(
                "simplify_topological"
            )
        postprocess["filter_background_modal_size"] = conf.predict.getint(
            "filter_background_modal_size"
        )
        query = conf.predict.get("reclassify_to_neighbour_query")
        if query is not None:
            query = query.replace("\n", " ")
        postprocess["reclassify_to_neighbour_query"] = query
        logger.info(f"Inline postprocessing:\n{pprint.pformat(postprocess)}")

        # Prepare the output dirs/paths
        predict_output_dir = Path(
            f"{conf.dirs.getpath('predict_image_output_basedir')!s}_"
            f"{predict_out_subdir}"
        )
        output_vector_dir = conf.dirs.getpath("output_vector_dir")
        output_vector_name = (
            f"{best_model['basefilename']}_{best_model['epoch']}_{image_layer}"
        )
        output_vector_path = output_vector_dir / f"{output_vector_name}.gpkg"

        # Start predict for entire dataset
        # --------------------------------
        # Send email
        email_helper.sendmail(
            f"Start predict for config {config_path.stem} on {image_layer}"
        )

        # Predict!
        nb_parallel = conf.general.getint("nb_parallel")
        predicter.predict_dir(
            model=model_for_predict,
            input_image_dir=input_image_dir,
            output_image_dir=predict_output_dir,
            output_vector_path=output_vector_path,
            classes=hyperparams.architecture.classes,
            min_probability=min_probability,
            postprocess=postprocess,
            border_pixels_to_ignore=conf.predict.getint("image_pixels_overlap"),
            projection_if_missing=image_layer_info["projection"],
            input_mask_dir=None,
            batch_size=batch_size,
            evaluate_mode=False,
            cancel_filepath=conf.files.getpath("cancel_filepath"),
            nb_parallel_postprocess=nb_parallel,
            max_prediction_errors=conf.predict.getint("max_prediction_errors"),
        )

        # Log and send mail
        message = f"Completed predict for config {config_path.stem}"
        logger.info(message)
        email_helper.sendmail(message)

        # Cleanup old data
        cleanup.clean_models(
            model_dir=conf.dirs.getpath("model_dir"),
            versions_to_retain=conf.cleanup.getint("model_versions_to_retain"),
            simulate=conf.cleanup.getboolean("simulate"),
        )
        cleanup.clean_training_data_directories(
            training_dir=conf.dirs.getpath("training_dir"),
            versions_to_retain=conf.cleanup.getint("training_versions_to_retain"),
            simulate=conf.cleanup.getboolean("simulate"),
        )
        cleanup.clean_predictions(
            output_vector_dir=conf.dirs.getpath("output_vector_dir"),
            versions_to_retain=conf.cleanup.getint("prediction_versions_to_retain"),
            simulate=conf.cleanup.getboolean("simulate"),
        )
    except Exception as ex:
        message = f"ERROR while running predict for task {config_path.stem}"
        logger.exception(message)
        email_helper.sendmail(
            subject=message, body=f"Exception: {ex}\n\n {traceback.format_exc()}"
        )
        raise Exception(message) from ex
    finally:
        if conf.tmp_dir is not None:
            shutil.rmtree(conf.tmp_dir, ignore_errors=True)


def main():
    """Run predict."""
    try:
        # Interprete arguments
        args = _predict_args(sys.argv[1:])

        # Run!
        predict(config_path=Path(args.config), config_overrules=args.config_overrules)
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise


# If the script is ran directly...
if __name__ == "__main__":
    main()



"""Module to download files from a github repo.

Based on https://github.com/sdushantha/gitdir/blob/master/gitdir/gitdir.py
"""
# mypy: ignore-errors

import json
import re
import ssl
import time
import urllib.request
from pathlib import Path
from typing import Optional, Union


def create_url(url: str) -> tuple[str, list[Path]]:
    """From the given url, produce a URL that is compatible with Github's REST API.

    Can handle blob or tree paths.

    Args:
        url (str): the base url.

    Returns:
        (str, list[Path]): Github REST API compatible url + directories to download.
    """
    # extract the branch name from the given url (e.g master)
    branch = re.findall(r"/tree/(.*?)/", url)
    api_url = url.replace("https://github.com", "https://api.github.com/repos")
    if len(branch) == 0:
        branch = re.findall(r"/blob/(.*?)/", url)[0]
        download_dirs = re.findall(r"/blob/" + branch + r"/(.*)", url)[0]
        api_url = re.sub(r"/blob/.*?/", "/contents/", api_url)
    else:
        branch = branch[0]
        download_dirs = re.findall(r"/tree/" + branch + r"/(.*)", url)[0]
        api_url = re.sub(r"/tree/.*?/", "/contents/", api_url)

    api_url = api_url + "?ref=" + branch
    return api_url, download_dirs


def download(
    repo_url: str,
    output_dir: Path,
    ssl_verify: Union[bool, str, None] = None,
    limit_rate: bool = True,
) -> Optional[int]:
    """Downloads the files and directories in repo_url.

    Args:
        repo_url (str): url to the repository to download.
        output_dir (Path): directory to download the repository to.
        ssl_verify (bool or str, optional): True or None to use the default
            certificate bundle as installed on your system. False disables
            certificate validation (NOT recommended!). If a path to a
            certificate bundle file (.pem) is passed, this will be used.
            In corporate networks using a proxy server this is often needed
            to evade CERTIFICATE_VERIFY_FAILED errors. Defaults to None.
        limit_rate (bool, optional): If True, limit the rate requests are done to
            github to the maximum level allowed for unauthenticated users.
            Defaults to True.
    """
    total_files = None
    context = None
    if ssl_verify is not None:
        # If it is a string, make sure it isn't actually a bool
        if isinstance(ssl_verify, str):
            if ssl_verify.lower() == "true":
                context = None
            elif ssl_verify.lower() == "false":
                ssl_verify = False
            else:
                context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                context.load_verify_locations(ssl_verify)

        if isinstance(ssl_verify, bool) and not ssl_verify:
            context = ssl._create_unverified_context()
            print("SSL VERIFICATION IS TURNED OFF!!!")

    # Generate the url which returns the JSON data
    api_url, download_dirs = create_url(repo_url)

    # To handle file names.
    if len(download_dirs.split(".")) == 0:
        dir_out = output_dir / download_dirs
    else:
        dir_out = output_dir / "/".join(download_dirs.split("/")[0:-1])

    # Download the file
    # If limit_rate enabled, always sleep 1 second before doing a request!
    if limit_rate:
        time.sleep(1)

    url = api_url
    try:
        with urllib.request.urlopen(url, context=context) as u:
            # Make a directory with the name which is taken from
            # the actual repo
            dir_out.mkdir(parents=True, exist_ok=True)

            # Total files count
            total_files = 0
            raw_data = u.read()
            data = json.loads(raw_data)

            # Get the total number of files
            total_files += len(data)

            # If the data is a file, download it as one.
            if isinstance(data, dict) and data["type"] == "file":
                # Download the file
                if limit_rate:
                    time.sleep(1)
                url = data["download_url"]
                with (
                    urllib.request.urlopen(url, context=context) as u,
                    open(dir_out / data["name"], "wb") as f,
                ):
                    f.write(u.read())

                return total_files

            # Loop over all files/dirs found
            for file in data:
                url = file["download_url"]
                path = output_dir / file["path"]
                path.parent.mkdir(parents=True, exist_ok=True)

                # If it is a file, download it, if dir, start recursively
                if url is not None:
                    if limit_rate:
                        time.sleep(1)
                    with (
                        urllib.request.urlopen(url, context=context) as u,
                        open(path, "wb") as f,
                    ):
                        f.write(u.read())
                else:
                    download(file["html_url"], output_dir)

    except urllib.error.HTTPError as ex:
        if ex.code == 403:
            message = f"Error: API Rate limit exceeded for url {url}"
        else:
            message = f"Error with url: {url}: {ex}"
        raise RuntimeError(message) from ex
    except Exception as ex:
        raise RuntimeError(f"Error with url: {url}: {ex}") from ex

    return total_files



"""Modile with generic Utility functions for vectorfile manipulations."""

import logging
from pathlib import Path

import geofileops as gfo

from orthoseg.util import vector_util

logger = logging.getLogger(__name__)


def reclassify_neighbours(
    input_path: Path,
    reclassify_column: str,
    query: str,
    output_path: Path,
    class_background: str = "background",
    force: bool = False,
):
    """Reclassify features to the class of neighbouring features.

    For features that comply to the query, if they have a neighbour (touch/overlap),
    change their classname to that of the neighbour with the longest intersection and
    dissolve to merge them.

    The query should follow the syntax of pandas.query queries and can us the following
    fields:
        - area (float): the area of the geometry.
        - perimeter (float): the perimeter of the geometry.

    Args:
        input_path (Path): input file path.
        reclassify_column (str): column to reclassify.
        query (str): th query to find the features to reclassify.
        output_path (Path): output file path.
        class_background (str, optional): the classname to treat as background.
            Defaults to "background".
        force (bool, optional): True to force calculation even if output file exists.
            Defaults to False.

    Raises:
        ValueError: raised if incompatible parameters are passed.
    """
    if output_path.exists():
        if not force:
            logger.info(
                f"reclassify_neighbours: return as output_path exists: {output_path}"
            )
            return
        gfo.remove(output_path)

    logger.info(f"reclassify_neighbours on {input_path} to {output_path}")
    input_gdf = gfo.read_file(input_path, columns=[reclassify_column])
    output_gdf = vector_util.reclassify_neighbours(
        input_gdf,
        reclassify_column=reclassify_column,
        query=query,
        border_bounds=None,
        class_background=class_background,
    )
    gfo.to_file(output_gdf, output_path)

    # Add/recalculate columns with area and nbcoords
    gfo.add_column(
        path=output_path,
        name="area",
        type=gfo.DataType.REAL,
        expression="ST_Area(geom)",
        force_update=True,
    )
    gfo.add_column(
        path=output_path,
        name="nbcoords",
        type=gfo.DataType.INTEGER,
        expression="ST_NPoints(geom)",
        force_update=True,
    )



from contextlib import nullcontext
from pathlib import Path

import pytest

from orthoseg import predict
from orthoseg.predict import _predict_args


@pytest.mark.parametrize(
    "args",
    [
        (
            [
                "--config",
                "X:/Monitoring/OrthoSeg/test/test.ini",
                "predict.image_layer=LT-2023",
            ]
        )
    ],
)
def test_predict_args(args):
    valid_args = _predict_args(args=args)
    assert valid_args is not None
    assert valid_args.config is not None
    assert valid_args.config_overrules is not None


@pytest.mark.parametrize("config_path, exp_error", [("INVALID", True)])
def test_predict(config_path, exp_error):
    if exp_error:
        handler = pytest.raises(ValueError)
    else:
        handler = nullcontext()
    with handler:
        predict(config_path=Path("INVALID"))
